from typing import Tuple

import pandas as pd
import numpy as np
from copy import deepcopy
import matplotlib.pyplot as plt
from IPython.display import display, HTML, display_html
import warnings

from lhcsmapi.Time import Time
from lhcsmapi.analysis.decorators import check_dataframe_empty
from lhcsmapi.analysis.diode.DiodeLeadResistanceAnalysis import DiodeLeadResistanceAnalysis
from lhcsmapi.analysis.diode.DiodeLeadResistancePlot import DiodeLeadResistancePlot
from lhcsmapi.analysis.expert_input import check_show_next, get_expert_comment
from lhcsmapi.pyedsl.AssertionBuilder import AssertionBuilder
from lhcsmapi.pyedsl.FeatureBuilder import FeatureBuilder
from lhcsmapi.pyedsl.PlotBuilder import PlotBuilder, create_title, convert_plot_features_map_to_text
from lhcsmapi.pyedsl.ResistanceBuilder import ResistanceBuilder
from lhcsmapi.reference.Reference import Reference
import lhcsmapi.pyedsl.SignalTransformationBuilder as SignalTransformationBuilder


def warning_on_one_line(message, category, filename, lineno, file=None, line=None):
    return '%s\n' % message


warnings.formatwarning = warning_on_one_line


def append(data_structure, col, values, suffixes=('',)):
    suffixes = ['']*len(values) if suffixes[0] == '' else suffixes
    for index, tau in enumerate(values):
        data_structure['%s%s' % (col, suffixes[index])] = tau


class CircuitAnalysis(object):

    def __init__(self, circuit_type, results_table, is_automatic=True, colormap='default'):
        self.circuit_type = circuit_type
        self.results_table = results_table
        self.is_automatic = is_automatic
        self.colormap = colormap

    # Timestamp table
    def create_timestamp_table(self, timestamp_dct, circuit_name=''):

        timestamp_dct_filtered = {}
        for key, value in timestamp_dct.items():
            if isinstance(value, pd.DataFrame):
                if not value.empty:
                    timestamp_dct_filtered[key] = value.loc[0, 'timestamp']
                else:
                    timestamp_dct_filtered[key] = 0
            else:
                timestamp_dct_filtered[key] = value if not np.isnan(value) else 0

        timestamp_df = pd.DataFrame({'Source': list(timestamp_dct_filtered.keys()),
                                     'Timestamp': list(timestamp_dct_filtered.values())})
        timestamp_df['Date and time'] = timestamp_df['Timestamp']. \
            apply(lambda col: Time.to_string_short(col) if col != 0 else 0)

        timestamp_df.sort_values(by='Timestamp', inplace=True)
        timestamp_df.reset_index(inplace=True, drop=True)

        # Put zero rows at the end
        zero_rows = timestamp_df.loc[timestamp_df['Timestamp'] == 0]
        non_zero_rows = timestamp_df.loc[timestamp_df['Timestamp'] != 0]
        timestamp_df = pd.concat([non_zero_rows, zero_rows])
        timestamp_df.reset_index(inplace=True, drop=True)

        # Compute time difference w.r.t. the first event
        timestamp_first_event = timestamp_df.loc[0, 'Timestamp']
        timestamp_df['dt w.r.t. origin [ms]'] = timestamp_df['Timestamp']. \
            apply(lambda col: (col - timestamp_first_event) * 1e-6 if col != 0 else 0)

        # Compute time difference w.r.t. the FGC
        if any(timestamp_df['Source'].str.contains('FGC')):
            timestamp_fgc = timestamp_df.loc[timestamp_df['Source'].str.contains('FGC'), 'Timestamp'].values[0]
            timestamp_df['dt w.r.t. FGC [ms]'] = timestamp_df['Timestamp']. \
                apply(lambda col: (col - timestamp_fgc) * 1e-6 if col != 0 else 0)

        # append the timestamps to the results table
        for k, v in timestamp_dct_filtered.items():
            if not (('iqps' in k.lower()) or ('nqps' in k.lower())):
                self.results_table['timestamp_%s' % k.lower()] = v

        display(HTML(timestamp_df.to_html()))

        if 'FGC' in timestamp_df.loc[0, 'Source']:
            warnings.warn('FGC is the first timestamp, potentially a PC trip not a quench!', stacklevel=2)

        if ('nQPS' in timestamp_dct_filtered) and ('iQPS' in timestamp_dct_filtered):
            if timestamp_dct_filtered['nQPS'] < timestamp_dct_filtered['iQPS']:
                warnings.warn('nQPS PM timestamp ({}) was created before iQPS one ({})'
                              .format(timestamp_dct_filtered['nQPS'], timestamp_dct_filtered['iQPS']), stacklevel=2)

        if 'nQPS' in timestamp_dct_filtered and any(timestamp_df['Source'].str.contains('FGC')):
            timestamp_fgc = timestamp_df.loc[timestamp_df['Source'].str.contains('FGC'), 'Timestamp'].values[0]
            if abs(timestamp_dct_filtered['nQPS'] - timestamp_fgc) > int(20e6):
                warnings.warn('nQPS and FGC PM events are not synchronized to +/- 20 ms!', stacklevel=2)

        if 'iQPS' in timestamp_dct_filtered and any(timestamp_df['Source'].str.contains('FGC')):
            timestamp_fgc = timestamp_df.loc[timestamp_df['Source'].str.contains('FGC'), 'Timestamp'].values[0]
            if abs(timestamp_dct_filtered['iQPS'] - timestamp_fgc) > int(20e6):
                warnings.warn('iQPS and FGC PM events are not synchronized to +/- 20 ms!', stacklevel=2)

    @staticmethod
    def create_ref_timestamp_table(timestamp_ref_dct):
        timestamp_ref_df = pd.DataFrame({'Source': list(timestamp_ref_dct.keys()),
                                         'Timestamp': list(timestamp_ref_dct.values())})
        timestamp_ref_df['Date and time'] = timestamp_ref_df['Timestamp'].apply(lambda col: Time.to_string_short(col))
        return timestamp_ref_df

    # PIC
    def analyze_pic(self, timestamp_pic):
        if any([np.isnan(timestamp_pic_el) for timestamp_pic_el in timestamp_pic]):
            warnings.warn('At least one timestamp is NaN, check of timestamp PIC difference is skipped.', stacklevel=2)
        else:
            dt_pic = abs(timestamp_pic[0] - timestamp_pic[1])
            if dt_pic > 1e6:
                warnings.warn('PIC timestamps (%s) and (%s) differ by more than 1 ms (%d ns)' %
                              (Time.to_string_short(timestamp_pic[0]), Time.to_string_short(timestamp_pic[1]), dt_pic))
            else:
                print('PIC timestamps (%s) and (%s) are less then 1 ms away.' %
                      (Time.to_string_short(timestamp_pic[0]), Time.to_string_short(timestamp_pic[1])))

        # Update results table
        if self.results_table is not None:
            self.results_table['%s_PIC' % self.circuit_type] = get_expert_comment('PIC', is_automatic=self.is_automatic)

    # PC
    @check_dataframe_empty(mode='any', warning='At least one DataFrame is empty, PC trigger analysis skipped!')
    def analyze_i_meas_pc_trigger(self, timestamp_fgc: int, events_action_df: pd.DataFrame,
                                  events_symbol_df: pd.DataFrame) -> None:
        """ Method finding and displaying the source of a PC trigger.
        If any input signal is missing, then the analysis is aborted.

        :param timestamp_fgc: FGC event timestamp
        :type timestamp_fgc: int
        :param events_action_df: PC events action
        :type events_action_df: pd.DataFrame
        :param events_symbol_df: PC events symbol
        :type events_symbol_df: pd.DataFrame
        """
        event_pc_df = pd.concat([events_action_df, events_symbol_df], axis=1)
        fault_event_pc_df = event_pc_df[event_pc_df['EVENTS.ACTION'] == 'FAULT']
        if not fault_event_pc_df.empty:
            array = np.asarray(fault_event_pc_df.index)
            idx = (np.abs(array - timestamp_fgc)).argmin()

            print('The reason for PC PM event is:')
            display(HTML(fault_event_pc_df[fault_event_pc_df.index == fault_event_pc_df.index[idx]].to_html()))

    @check_dataframe_empty(mode='all', warning='I_MEAS signal is empty, dI_MEAS/dt analys skipped!')
    def calculate_current_slope(self, i_meas_df: pd.DataFrame, col_name=['Ramp Rate', 'Plateau Duration']) -> None:
        """ Method analyzing the power converter current derivative.
        If the input signal is missing, then the analysis is skipped.
        In case the current derivative prior to an event was zero (less than 0.1 A/s), then
        - the value of the current plateau and its duration are reported in the results table
        otherwise only the value of the derivative (ramp-rate) is reported in the results table

        :param i_meas_df: measured power converter current
        :type i_meas_df: pd.DataFrame
        :param col_name: name of a column to set in results table
        :type col_name: str
        """
        # Ramp Rate and Plateau Duration
        pd.options.mode.chained_assignment = None
        idx = max([i_meas_df.idxmin().values[0], i_meas_df.idxmax().values[0]])
        i_meas_df_der = i_meas_df[i_meas_df.index < idx]

        last_di_dt, last_di_dt_duration = SignalTransformationBuilder.calculate_last_const_di_dt(i_meas_df_der)
        print('Last di/dt prior to an FPA is %f A/s.' % last_di_dt)
        print('Duration of the last di/dt prior to an FPA is %f s (as seen in the PM buffer).' % last_di_dt_duration)
        # Store analysis result
        if self.results_table is not None:
            if abs(last_di_dt) > 0.1:
                self.results_table[col_name[0]] = last_di_dt
                self.results_table[col_name[1]] = 0
            else:
                self.results_table[col_name[0]] = 0
                self.results_table[col_name[1]] = last_di_dt_duration

        pd.options.mode.chained_assignment = 'warn'

    @check_dataframe_empty(mode='any', warning='At least one DataFrame is empty, PC trigger analysis skipped!')
    def calculate_current_miits(self, i_meas_df: pd.DataFrame, t_quench: float, col_name='I_MEAS MIITs') -> None:
        """ Method calculating quench load (MIITs) of the power converter current.
        If any input signal is missing, then the analysis is aborted.
        The calculated MIITs is set in the results table.

        :param i_meas_df: measured and filtered power converter current
        :type i_meas_df: pd.DataFrame
        :param t_quench: time of quench start (in seconds)
        :type t_quench: pd.DataFrame
        :param col_name: name of a column to set the output results table
        :type col_name: str
        """
        i_meas_pos_df = i_meas_df[i_meas_df.index >= t_quench]
        miits_i_meas = FeatureBuilder().with_signal(i_meas_pos_df).calculate_features('miits').loc[0, :].values[0]
        print('%s is %s MA^2s.' % (col_name, miits_i_meas))
        # Store analysis result
        if self.results_table is not None:
            self.results_table[col_name] = miits_i_meas

    @check_dataframe_empty(mode='any', warning='PC signals are empty, plot is skipped!')
    def plot_i_meas_pc(self, circuit_name: str, timestamp_fgc: int, i_dfs, i_ref_dfs, xlim=(np.nan, np.nan),
                       ylim=(np.nan, np.nan)) -> None:
        """ Method plotting power converter currents.
        If any input signal is missing, then the analysis is skipped.

        :param circuit_name: name of the analyzed circuit (for the title)
        :type circuit_name: str
        :param timestamp_fgc: timestamp of the FGC PM event (for the title)
        :type timestamp_fgc: int
        :param i_dfs: list of power converter currents
        :type i_dfs: List[pd.DataFrame]
        :param i_ref_dfs: list of reference power converter currents
        :type i_ref_dfs: List[pd.DataFrame]
        :param xlim: limits of the x-axis
        :type xlim: Tuple[float]
        :param ylim: limits of the y-axis
        :type ylim: Tuple[float]
        """
        title = create_title(circuit_name, timestamp_fgc, [i_df.columns[0] for i_df in i_dfs])

        # Global view
        PlotBuilder().with_signal(i_dfs, title=title, figsize=(20, 7), grid=True) \
            .with_ylabel(ylabel='I, [A]') \
            .with_signal(i_ref_dfs) \
            .with_ylabel(ylabel='I_REF, [A]') \
            .with_xlim(lim=xlim) \
            .with_ylim(lim=ylim) \
            .plot()

    @check_dataframe_empty(mode='any', warning='PC signals are empty, zoomed plot is skipped!')
    def plot_i_meas_pc_zoom(self, circuit_name: str, timestamp_fgc: int, t_quench: float, i_dfs, i_ref_dfs,
                            xlim=(np.nan, np.nan), ylim=(np.nan, np.nan)) -> None:
        """ Method plotting power converter currents in zoom view around the start of a quench or
        If any input signal is missing, then the analysis is skipped.

        :param circuit_name: name of the analyzed circuit (for the title)
        :type circuit_name: str
        :param timestamp_fgc: timestamp of the FGC PM event (for the title)
        :type timestamp_fgc: int
        :param t_quench: time of quench start (in seconds)
        :type t_quench: pd.DataFrame
        :param i_dfs: list power converter currents
        :type i_dfs: List[pd.DataFrame]
        :param i_ref_dfs: list power converter currents
        :type i_ref_dfs: List[pd.DataFrame]
        :param xlim: limits of the x-axis
        :type xlim: Tuple[float]
        :param ylim: limits of the y-axis
        :type ylim: Tuple[float]
        """
        title = create_title(circuit_name, timestamp_fgc, [i_df.columns[0] for i_df in i_dfs])

        # Zoomed view
        # # in x
        if np.isnan(xlim[0]) and np.isnan(xlim[1]):
            xlim = [t_quench - 0.01, 0.01]

        # # in y
        if np.isnan(ylim[0]) and np.isnan(ylim[1]):
            mid = 0
            for i_df in i_dfs:
                idx = i_df.index.get_loc(t_quench, 'nearest')
                mid = max(mid, i_df[i_df.index == i_df.index[idx]].values[0][0])
            ylim = [mid - 10, mid + 10]

        PlotBuilder().with_signal(i_dfs, title=title, figsize=(15, 15), grid=True) \
            .with_ylabel(ylabel='I, [A]').with_ylim(lim=ylim) \
            .with_axvline(x=t_quench, ls='--') \
            .with_signal(i_ref_dfs, title=title, figsize=(15, 15), grid=True) \
            .with_ylabel(ylabel='I_REF, [A]').with_ylim(lim=ylim) \
            .with_xlim(lim=xlim) \
            .plot()

    @check_dataframe_empty(mode='any', warning='At least one DataFrame is empty, I_MEAS quench analysis skipped!')
    def calculate_quench_current(self, i_meas_df: pd.DataFrame, t_quench: float,
                                         col_name='Quench Current') -> None:
        """ Method calculating the quench current, i.e., the current at the moment when I_A and I_REF start to
        deviate from one another.
        If any input signal is missing, then the analysis is skipped.
        The calculated quench current is set in the results table.

        :param i_meas_df: measured power converter current
        :type i_meas_df: pd.DataFrame
        :param t_quench: time of quench start (in seconds)
        :type t_quench: pd.DataFrame
        :param col_name: name of a column to set in results table
        :type col_name: str
        """
        i_meas_pos_df = i_meas_df[i_meas_df.index >= t_quench]
        print('%s is %s A.' % (col_name, round(i_meas_pos_df.iloc[0].values[0], 1)))

        # Store analysis result
        if self.results_table is not None:
            self.results_table[col_name] = round(i_meas_pos_df.iloc[0].values[0], 1)

    # Primarily for RB and RQ
    @check_dataframe_empty(mode='any', warning='PC signals are empty, I_MEAS plot is skipped!')
    def analyze_i_meas_pc(self, circuit_name, timestamp_fgc, timestamp_fgc_ref, timestamp_pic,
                          i_meas_df, i_meas_ref_df, duration_decay=(2, 100)):
        # Calculate Features
        # # characteristic time constant
        tau_i_meas, tau_i_meas_ref = FeatureBuilder().with_signal(i_meas_df, i_meas_ref_df) \
            .subtract_min_value() \
            .with_duration(t_start=duration_decay[0], t_end=duration_decay[1]) \
            .calculate_features('tau_energy') \
            .values[0]

        # Plot
        title = '{} {}: I_MEAS(t) and I_MEAS(t)/dI_MEAS(t) (+PNO.b3)\n'.\
            format(Time.to_string_short(timestamp_fgc), circuit_name)

        tau_i_meas_df = SignalTransformationBuilder.calculate_char_time_profile(i_meas_df)
        tau_i_meas_ref_df = SignalTransformationBuilder.calculate_char_time_profile(i_meas_ref_df)

        # Plot
        plot_features = {'Circuit name': [circuit_name, ''],
                         'PIC Abort': [Time.to_string_short(timestamp_pic), ''],
                         'PC Off': [Time.to_string_short(timestamp_fgc), ''],
                         'PNO.b3': [Time.to_string_short(timestamp_fgc_ref), ''],
                         'max(I_MEAS)': [round(i_meas_df.max().values[0]), 'A'],
                         'max(I_MEAS, PNO.b3)': [round(i_meas_ref_df.max().values[0]), 'A'],
                         'I_MEAS: tau': [round(tau_i_meas), 's'],
                         'I_MEAS(PNO.b3):tau': [round(tau_i_meas_ref), 's']}
        plot_text = convert_plot_features_map_to_text(plot_features)

        hwc_test = 'PNO.b2' if self.circuit_type == 'RB' else 'PNO.b3'
        ylim = (50, 110) if self.circuit_type == 'RB' else (0, 35)
        ax2 = PlotBuilder().with_signal([i_meas_df, i_meas_ref_df], title=title, grid=True, figsize=(15, 10),
                                        style=['-', '--']) \
                .with_ylabel(ylabel='I_MEAS, [A]') \
                .with_legend(labels=['I_MEAS', 'I_MEAS (%s)' % hwc_test]) \
                .with_signal([tau_i_meas_df, tau_i_meas_ref_df], style=['-', '--']) \
                .with_ylabel(ylabel='-I_MEAS/dI_MEAS/dt [s]')\
                .with_legend(labels=['-I_MEAS/dI_MEAS_dt', '-I_MEAS/dI_MEAS_dt (%s)' % hwc_test]) \
                .with_ylim(ylim).with_xlim((-50, np.nan)) \
                .plot(show_plot=False).get_axes()[1]

        ax2.annotate(plot_text, xytext=(ax2.get_xlim()[1] * 1.2, ax2.get_ylim()[0]), xy=(20, ax2.get_ylim()[0]),
                     fontsize=15, family='monospace')
        plt.show()

        # Compare features to the reference
        pc_act_df = pd.DataFrame({'tau_i_meas': tau_i_meas}, index=['act'])

        tau_i_meas_comp_df = Reference.compare_features_to_reference(pc_act_df, self.circuit_type, 'PC')

        display(HTML(tau_i_meas_comp_df.to_html()))

        is_tau_i_meas_correct = Reference.check_feature_comparison(tau_i_meas_comp_df)

        # Update results table
        if self.results_table is not None:
            status = 'Within reference' if is_tau_i_meas_correct else 'Outside of reference'
            analysis = '%s_PC_I_MEAS_tau' % circuit_name.split('.')[0]
            self.results_table[analysis] = get_expert_comment(analysis, self.is_automatic, status)

    # PC - Earth Current
    @check_dataframe_empty(mode='any', warning='I_EARTH signal is empty, analysis skipped!')
    def calculate_max_i_earth_pc(self, i_earth_df: pd.DataFrame, col_name='Earth Current') -> None:
        """ Method calculating the absolute maximum value of earth current.
        If the input signal is missing, then the calculation is skipped.
        The calculated absolute maximum value is set in the results table.

        :param i_earth_df: current of the earth circuit in the power converter
        :type i_earth_df: pd.DataFrame
        :param col_name: name of a column to set in results table
        :type col_name: str
        """

        i_earth_max = FeatureBuilder().with_signal(i_earth_df).calculate_features('max_abs_min').loc[0, :].values[0]
        print('The maximum %s is %s A.' % (col_name, round(i_earth_max, 3)))

        # Store analysis result
        if self.results_table is not None:
            self.results_table[col_name] = round(i_earth_max, 3)

    @check_dataframe_empty(mode='any', warning='I_EARTH signal is empty, analysis skipped!')
    def plot_i_earth_pc(self, circuit_name: str, timestamp_fgc: int, i_earth_df: pd.DataFrame) -> None:
        """ Method plotting the earth current.
        If the input signal is missing, then the plot is skipped.

        :param circuit_name: name of the analyzed circuit (for the title)
        :type circuit_name: str
        :param timestamp_fgc: timestamp of the FGC PM event (for the title)
        :type timestamp_fgc: int
        :param i_earth_df: current of the earth circuit in the power converter
        :type i_earth_df: pd.DataFrame
        """
        title = create_title(circuit_name, timestamp_fgc, 'I_EARTH')
        PlotBuilder().with_signal(i_earth_df, title=title, grid=True) \
            .with_ylabel(ylabel='IEARTH [A]') \
            .plot()

    @check_dataframe_empty(mode='any', warning='At least one DataFrame is empty, I_EARTH analysis skipped!')
    def analyze_i_earth_pc(self, circuit_type, circuit_name, timestamp_fgc, i_a_df, i_earth_df,
                           i_earth_ref_df, xlim=(-0.1, 0.3)):

        hwc_test = 'PNO.b2' if circuit_type == 'RB' else 'PNO.b3'
        title = '{} {}: I_A and I_EARTH (+{})'.format(Time.to_string_short(timestamp_fgc), circuit_name, hwc_test)

        PlotBuilder().with_signal(i_a_df, title=title, grid=True) \
            .with_ylabel(ylabel='I_A, [A]') \
            .with_signal([i_earth_df, i_earth_ref_df], style=['-', '--']) \
            .with_ylabel(ylabel='I_EARTH, [A]') \
            .with_legend(labels=['I_EARTH', 'I_EARTH (%s)' % hwc_test]) \
            .with_xlim(xlim) \
            .plot()

        # Update results table
        if self.results_table is not None:
            analysis = '%s_PC_I_EARTH' % circuit_name.split('.')[0]
            self.results_table[analysis] = get_expert_comment(analysis, self.is_automatic)

    @check_dataframe_empty(mode='any', warning='At least one DataFrame is empty, I_EARTH_PCNT analysis skipped!')
    def analyze_i_earth_pcnt_pc(self, circuit_type: str, circuit_name: str, timestamp_fgc: int, i_meas_df: pd.DataFrame,
                                i_meas_ref_df: pd.DataFrame, i_earth_pcnt_df: pd.DataFrame,
                                i_earth_pcnt_ref_df: pd.DataFrame, xlim=(5, 120)):

        hwc_test = 'PNO.b2' if circuit_type == 'RB' else 'PNO.b3'
        title = '{} {}: I_MEAS and I_EARTH (+{})'.format(Time.to_string_short(timestamp_fgc), circuit_name, hwc_test)
        scaling = (i_meas_df.max() / i_meas_ref_df.max()).values[0]
        PlotBuilder().with_signal(i_meas_df, title=title, grid=True) \
            .with_ylabel(ylabel='I_MEAS, [A]') \
            .with_signal([i_earth_pcnt_df[i_earth_pcnt_df.index>3],
                          i_earth_pcnt_ref_df[i_earth_pcnt_ref_df.index > 3]*scaling], yerr=[None, 3]) \
            .with_ylabel(ylabel='I_EARTH_PCNT, [%]') \
            .with_legend(labels=['I_EARTH_PCNT', 'I_EARTH_PCNT (%s)' % hwc_test]) \
            .with_xlim(xlim) \
            .plot()

        is_i_earth_pcnt_ok = AssertionBuilder().with_signal(i_earth_pcnt_df[i_earth_pcnt_ref_df.index > 3]) \
            .is_between_two_references(first_ref=i_earth_pcnt_ref_df[i_earth_pcnt_ref_df.index > 3],
                                       first_scaling=scaling, first_offset=3,
                                       second_ref=i_earth_pcnt_ref_df[i_earth_pcnt_ref_df.index > 3],
                                       second_scaling=scaling, second_offset=3, is_plot=False) \
            .get_assertion_result()

        # Update results table
        if self.results_table is not None:
            analysis = '%s_PC_I_EARTH_PCNT' % circuit_name.split('.')[0]
            status = 'Within reference' if is_i_earth_pcnt_ok else 'Outside of reference'
            self.results_table[analysis] = get_expert_comment(analysis, self.is_automatic, status)

    # EE
    @check_dataframe_empty(mode='any', warning='At least one DataFrame is empty, tau_U_DUMP_RES analysis skipped!')
    def analyze_char_time_u_dump_res_ee(self, circuit_name, timestamp_fgc, u_dump_res_dfs, i_meas_df,
                                        duration_decay=(1, 120)):
        # Suffixes
        suffixes = ('_odd', '_even') if self.circuit_type == 'RB' else ('',)

        # Vectorize Input
        u_dump_res_dfs = u_dump_res_dfs if isinstance(u_dump_res_dfs, list) else [u_dump_res_dfs]

        # Calculate Features
        # # characteristic time constant
        tau_u_dump_res = FeatureBuilder().with_signal(u_dump_res_dfs) \
            .subtract_min_value() \
            .with_duration(t_start=duration_decay[0], t_end=duration_decay[1]) \
            .calculate_features('tau_energy') \
            .values[0]

        # # resistance
        def calculate_ee_resistance(u_dump_res_df, i_meas_df):
            index_i_meas_u_dump_res_max = i_meas_df.index.get_loc(u_dump_res_df.idxmax().values[0], method='nearest')
            i_meas_u_dump_res_max = i_meas_df.iloc[index_i_meas_u_dump_res_max].values[0]
            return u_dump_res_df.max().values[0] / i_meas_u_dump_res_max

        r_dump_res = [calculate_ee_resistance(u_dump_res_df, i_meas_df) for u_dump_res_df in u_dump_res_dfs]

        # Plot
        title = '{} {} EE: I_MEAS(t) and U_DUMP_RES(t)'.format(Time.to_string_short(timestamp_fgc), circuit_name)

        plot_features = {'Circuit name': [circuit_name, '']}
        tau_u_dump_res_plot = [(round(tau), 's') for tau in tau_u_dump_res]
        append(plot_features, 'tau: EE', tau_u_dump_res_plot, suffixes=suffixes)

        r_dump_res_plot = [(round(res * 1000, 3), 'mOhm') for res in r_dump_res]
        append(plot_features, 'res: EE', r_dump_res_plot, suffixes=suffixes)

        plot_text = convert_plot_features_map_to_text(plot_features)

        PlotBuilder().with_signal(i_meas_df, title=title, grid=True) \
            .with_ylabel(ylabel='I_MEAS, [A]') \
            .with_signal(u_dump_res_dfs, color=['C2', 'C3']) \
            .with_ylabel(ylabel='U_DUMP_RES, [V]') \
            .with_xlim((-50, np.nan)) \
            .with_annotate(s=plot_text, xy=(0.7, 0),
                           xytext=(max(u_dump_res_dfs[0].index.max() * 1.3, i_meas_df.index.max()) * 1.1, 0)) \
            .plot()

        # Update results table
        # Compare features to the reference
        ee_act_df = pd.DataFrame(index=['act'])
        append(ee_act_df, 'tau_u_dump_res', tau_u_dump_res, suffixes=suffixes)
        append(ee_act_df, 'R', r_dump_res, suffixes=suffixes)

        res_tau_u_dump_res_comp_df = Reference.compare_features_to_reference(ee_act_df, self.circuit_type, 'EE')
        is_res_tau_u_dump_res_correct = Reference.check_feature_comparison(res_tau_u_dump_res_comp_df)

        display(HTML(res_tau_u_dump_res_comp_df.to_html()))

        if self.results_table is not None:
            status = 'Within threshold(s)' if is_res_tau_u_dump_res_correct else 'Outside of threshold(s)'
            analysis = '%s_EE_U_DUMP_RES_tau_res' % circuit_name.split('.')[0]
            self.results_table[analysis] = get_expert_comment(analysis, self.is_automatic, status)

    @check_dataframe_empty(mode='any', warning='At least one DataFrame is empty, t_delay_U_DUMP_RES analysis skipped!')
    def analyze_delay_time_u_dump_res_ee(self, circuit_name, timestamp_fgc, timestamp_pic, timestamp_ees,
                                         i_a_df, i_ref_df, u_dump_res_dfs):
        # Suffixes
        suffixes = ('_odd', '_even') if self.circuit_type == 'RB' else ('',)

        # Vectorize Input
        u_dump_res_dfs = u_dump_res_dfs if isinstance(u_dump_res_dfs, list) else [u_dump_res_dfs]
        timestamp_ees = timestamp_ees if isinstance(timestamp_ees, list) else [timestamp_ees]

        # Calculate Features
        t_delay_ees = [(timestamp_ee - timestamp_pic) * 1e-9 for timestamp_ee in timestamp_ees]

        ee_act_df = pd.DataFrame(index=['act'])
        append(ee_act_df, 't_delay_ee', t_delay_ees, suffixes=suffixes)

        # Plot
        plot_features = {'Circuit name': [circuit_name, '']}

        # # add calculated features
        t_delay_ees_ms = [(round(t_delay_ee * 1e3), 'ms') for t_delay_ee in t_delay_ees]
        append(plot_features, 't_delay_ee', t_delay_ees_ms, suffixes=suffixes)

        # Plot
        plot_text = convert_plot_features_map_to_text(plot_features)
        title = '{} {} EE: I_A(t) and U_DUMP_RES(t)'.format(Time.to_string_short(timestamp_fgc), circuit_name)
        if all(i_a_df.index < 1):
            ax1_ylim_min = int(i_a_df.values[-1])
            ax1_xlim_min = i_a_df.index[0]
        else:
            ax1_ylim_min = int(i_a_df.reindex([1], method='nearest', tolerance=.1).values[0])
            ax1_xlim_min = -0.2

        if all(i_a_df.index < 1):
            ax1_ylim_max = (i_a_df.values[0])
            ax1_xlim_max = i_a_df.index[-1]
        else:
            ax1_ylim_max = int(i_a_df.reindex([-0.1], method='nearest', tolerance=.1).values[0]) * 1.001
            ax1_xlim_max = 1

        ax2 = PlotBuilder().with_signal([i_a_df, i_ref_df], title=title, grid=True, marker=[None, 'o']) \
                .with_ylabel(ylabel='I_A, [A]') \
                .with_xlim((ax1_xlim_min, ax1_xlim_max)).with_ylim((ax1_ylim_min, ax1_ylim_max)) \
                .with_signal(u_dump_res_dfs, color=['C2', 'C3']) \
                .with_ylabel(ylabel='U_DUMP_RES, [V]') \
                .with_xlim((ax1_xlim_min, ax1_xlim_max)) \
                .with_annotate(s=plot_text, xy=(0.6, 0), xytext=(1.1, 0)) \
                .with_axvline(x=(timestamp_pic - timestamp_fgc) * 1e-9, color='green', linestyle='--', linewidth=1.5) \
                .plot(show_plot=False).get_axes()[1]

        for index, timestamp_ee in enumerate(timestamp_ees):
            ax2.annotate('', xy=((timestamp_ee - timestamp_fgc) * 1e-9, (0.2 - index * 0.1) * ax2.get_ylim()[1]),
                         xytext=((timestamp_pic - timestamp_fgc) * 1e-9, (0.2 - index * 0.1) * ax2.get_ylim()[1]),
                         arrowprops=dict(arrowstyle='<->'))
        plt.show()

        # Compare features to the reference
        t_delay_ee_comp_df = Reference.compare_features_to_reference(ee_act_df, self.circuit_type, 'EE',
                                                                        timestamp=timestamp_fgc)
        is_t_delay_ee_correct = Reference.check_feature_comparison(t_delay_ee_comp_df)

        display(HTML(t_delay_ee_comp_df.to_html()))

        # Update results table
        if self.results_table is not None:
            status = 'Within threshold(s)' if is_t_delay_ee_correct else 'Outside of threshold(s)'
            analysis = '%s_EE_U_DUMP_RES_delay' % circuit_name.split('.')[0]
            self.results_table[analysis] = get_expert_comment(analysis, self.is_automatic, status)

    def analyze_ee_temp(self, circuit_name, timestamp_ee, t_res_act_dfs, t_res_ref_dfs, *, abs_margin=25, scaling=1):
        title = create_title(circuit_name, timestamp_ee, 'T_RES_BODY')
        is_signal_correct = AssertionBuilder().with_signal(t_res_act_dfs) \
            .compare_to_reference(signal_ref_dfs=t_res_ref_dfs, abs_margin=abs_margin, scaling=scaling) \
            .show_plot(ylabel='T_RES_BODY, [degC]', title=title) \
            .get_assertion_result()

        if self.results_table is not None:
            status = 'Within reference' if is_signal_correct else 'Outside of reference'
            analysis = '%s_EE_T_RES_BODY' % circuit_name.split('.')[0]
            self.results_table[analysis] = get_expert_comment(analysis, self.is_automatic, status)

    # Quench
    @staticmethod
    def find_time_of_quench(i_ref_df: pd.DataFrame) -> float:
        """ Method finding the start of a quench as one time stamp before the drop of I_REF to 0 A around t = 0 s;
        +/- 2 s.

        :param i_ref_df: reference current of the power converter
        :type i_ref_df: pd.DataFrame
        :return: one time stamp before the drop of I_REF to 0 A
        """
        t_i_ref_min = i_ref_df[(i_ref_df.index > -2) & (i_ref_df.index < 2)].idxmin().values[0]
        return i_ref_df[i_ref_df.index < t_i_ref_min].index[-1]

    # QDS
    @check_dataframe_empty(mode='all', warning='U_RES signal is empty, analysis skipped!')
    def calculate_u_res_slope(self, u_res_df: pd.DataFrame, col_name='dU_RES_dt') -> pd.DataFrame:
        """ Method calculating and plotting the slope of the resistive voltage increase
        If any input signal is missing, then the analysis is skipped.
        The calculated voltage derivative is set in the results table.

        :param circuit_name: name of the analyzed circuit (for the title)
        :type circuit_name: str
        :param timestamp_qps: timestamp of the QPS PM event (for the title)
        :type timestamp_qps: int
        :param u_res_df: resistive voltage signal
        :type u_res_df: pd.DataFrame
        :param col_name: name of a column to set the output results table
        :type col_name: str
        """
        # ToDo duplication of what already exist!
        # Skip first ten points to avoid initial noise
        u_res_slice_df = u_res_df[u_res_df.index > u_res_df.index[10]]
        # Take absolute value
        u_res_flip_df = SignalTransformationBuilder.flip_u_qs0_signal(u_res_slice_df)

        # Find first index of argmax (0.2 V) - if the threshold was not hit, return an empty DataFrame
        if u_res_flip_df[u_res_flip_df >= 0.2].dropna().empty:
            return pd.DataFrame(data=[0], index=[0], columns=['U_RES_slope'])

        t_u_max = u_res_flip_df[u_res_flip_df >= 0.2].dropna().index[0]
        v_u_max = u_res_df[u_res_df.index == t_u_max].dropna().values[0]

        # Find index of voltage equal to 0.025 V
        temp = u_res_flip_df[(u_res_flip_df.index < t_u_max)]
        t_u_min = temp[(temp.values < 0.025)].index[-1]
        v_u_min = u_res_df[u_res_df.index == t_u_min].dropna().values[0][0]

        # Calculate dU_RES/dt
        d_u_res_dt = (v_u_max - v_u_min) / (t_u_max - t_u_min)

        # Store analysis result
        if self.results_table is not None:
            self.results_table[col_name] = d_u_res_dt[0]

        return pd.DataFrame(data=[v_u_min, v_u_max[0]], index=[t_u_min, t_u_max], columns=['U_RES_slope'])

    @staticmethod
    @check_dataframe_empty(mode='any', warning='At least one DataFrame is empty, plot of U_RES, U_1, U2 skipped!')
    def plot_u_res_u_res_slope_u_1_u_2(circuit_name, timestamp_qds, u_res_df, u_res_slope_df, u_1_df, u_2_df,
                                       suffix=''):
        """ Method plotting U_RES, its slope along with U_1 and U_2 signals used for quench detection.

        :param circuit_name: name of the analyzed circuit (for the title)
        :type circuit_name: str
        :param timestamp_qds: timestamp of the QDS PM event (for the title)
        :type timestamp_qds: int
        :param u_res_df: quench detection voltage
        :type u_res_df: pd.DataFrame
        :param u_res_slope_df: initial slope of the voltage
        :type u_res_slope_df: pd.DataFrame
        :param u_1_df: voltage of the first aperture used for comparison
        :type u_1_df: pd.DataFrame
        :param u_2_df: voltage of the second aperture used for comparison
        :type u_2_df: pd.DataFrame
        :param suffix: magnet name, e.g., '_Q1'
        """
        title = create_title(circuit_name, timestamp_qds,
                             ['U_RES' + suffix, 'U_RES_slope' + suffix, 'U_1' + suffix, 'U_2' + suffix])
        PlotBuilder().with_signal([u_res_df, u_res_slope_df], title=title, figsize=(15, 7), grid=True) \
            .with_ylabel(ylabel='U_RES, [V]') \
            .with_xlim((-0.5, 0.5)) \
            .with_signal([u_1_df, u_2_df]) \
            .with_ylabel(ylabel='U_1, U_2, [V]') \
            .plot()

    # NQPS Voltage: U_DIODE, U_EARTH
    @check_dataframe_empty(mode='all', warning='All DataFrames are empty, nQPS voltage plot skipped!')
    def filter_quenched_magnets(self, u_diode_rb_dfs):
        def make_filter_quenched_magnets(quenched_magnets):
            def name_filter(df):
                return any([magnet in df.columns[0] for magnet in quenched_magnets])

            return name_filter

        filter_quenched_magnets = make_filter_quenched_magnets(self.results_table['magnet'].values)

        return list(filter(filter_quenched_magnets, u_diode_rb_dfs))

    @check_dataframe_empty(mode='all', warning='All DataFrames are empty, nQPS voltage plot skipped!')
    def analyze_u_diode_nqps(self, circuit_name, timestamp_fgc, i_meas_df, u_dfs, signal, system, legend=False,
                             xlim=(-5, 150)):

        title = '%s %s I_MEAS(t) and %s(t)' % (Time.to_string_short(timestamp_fgc), circuit_name, signal)
        PlotBuilder().with_signal(i_meas_df, title=title, grid=True) \
            .with_ylabel(ylabel='I_MEAS, [A]') \
            .with_signal(u_dfs, legend=legend) \
            .with_ylabel(ylabel='%s, [V]' % signal) \
            .with_xlim(xlim) \
            .plot()

        if self.results_table is not None:
            system_signal = '{}_{}'.format(system, signal)
            self.results_table[system_signal] = get_expert_comment(system_signal, self.is_automatic)

    # Quench Heaters
    def analyze_qh(self, circuit_name, timestamp_qds, u_hds_dfs, u_hds_ref_dfs):
        """ Method analyzing QH voltage signals in IT circuit. It is plotting all signals and calculating:
            - initial voltage
            - final voltage
            - time constant based on the charge approach

        :param circuit_name: circuit name to analyze ('RQX.R1')
        :type circuit_name: str
        :param u_hds_dfs: list of lists of QH voltage signals. There are as many outer lists as quenched
            magnets and as many elements in the inner list as there are QH voltage signals to analyze
        :type u_hds_dfs: List[List[pd.DataFrame]]
        :param u_hds_ref_dfs: list of lists of reference QH voltage signals. There are as many outer lists as
            quenched magnets and as many elements in the inner list as there are reference QH voltage signals to analyze
        :type u_hds_ref_dfs: List[List[pd.DataFrame]]
        """

        # Plot
        title = create_title(circuit_name, timestamp_qds, 'U_HDS')
        PlotBuilder().with_signal(u_hds_dfs, title=title, figsize=(15, 7), grid=True) \
            .with_xlim((-0.5, 1.5)) \
            .with_signal(u_hds_ref_dfs, is_twinx=False, grid=True, style='--') \
            .with_ylabel(ylabel='U_HDS, [V]', color='C0') \
            .plot()

        PlotBuilder().with_signal(u_hds_dfs, title=title, figsize=(15, 7), grid=True) \
            .with_xlim((-0.5, 1.5)) \
            .with_signal(u_hds_ref_dfs, is_twinx=False, grid=True, style='--') \
            .with_ylabel(ylabel='U_HDS, [V]', color='C0') \
            .plot(yscale='log')

        # Take decay only
        u_hds_decay_dfs = [u_hds_df[u_hds_df.index >= 0] for u_hds_df in u_hds_dfs]
        u_hds_ref_decay_dfs = [u_hds_ref_df[u_hds_ref_df.index >= 0] for u_hds_ref_df in u_hds_ref_dfs]

        # Calculate features
        # # compare to reference values
        first_last_u_df = FeatureBuilder().with_signal(u_hds_decay_dfs) \
            .calculate_features(features=['first', 'last20mean'], index=timestamp_qds)

        first_last_u_comp_df = Reference.compare_features_to_reference(first_last_u_df, self.circuit_type, 'QH',
                                                                       wildcard={'CIRCUIT': circuit_name})
        is_first_last_u_ok = Reference.check_feature_comparison(first_last_u_comp_df)

        # # compare to a reference discharge
        timestamp_ref = Reference.get_quench_heater_reference_discharge(self.circuit_type, circuit_name)

        tau_df = FeatureBuilder().with_signal(u_hds_decay_dfs) \
            .calculate_features(features=['tau_charge'], index=timestamp_qds)

        tau_ref_df = FeatureBuilder().with_signal(u_hds_ref_decay_dfs) \
            .calculate_features(features=['tau_charge'], index=timestamp_ref)

        tau_comp_df = Reference.compare_difference_of_features_to_reference(tau_df, tau_ref_df, self.circuit_type,
                                                                            'QH',
                                                                            wildcard={'CIRCUIT': circuit_name})
        is_tau_correct = Reference.check_feature_comparison(tau_comp_df)

        # Update results table
        # Use OK rather than True always
        final_label = 'OK' if is_first_last_u_ok and is_tau_correct else 'NOT_OK'
        if self.results_table is not None:
            self.results_table['QH'] = final_label

        # Display features
        # Display comparison
        first_last_u_comp_styler = first_last_u_comp_df.style.set_table_attributes(
            "style='display:inline'").set_caption('Comparison of the inititial and final voltage')
        tau_comp_styler = tau_comp_df.style.set_table_attributes("style='display:inline'").set_caption(
            'Comparison of the discharge characteristic time to the reference')

        display_html(first_last_u_comp_styler._repr_html_() + tau_comp_styler._repr_html_(), raw=True)

        print('The QH discharges are labeled: %s.' % final_label)

    # DIODE
    def analyze_diode_leads(self, circuit_name, i_meas_u_diode_u_ref_pm_dfs, i_meas_u_diode_nxcals_dfs):
        index_max = self.results_table.index[-1]
        for index, row in self.results_table.iterrows():
            current_col = 'I_MEAS_quench' if self.circuit_type == 'RB' else 'I_MEAS_RQD_quench'
            source_qds, timestamp_qds, u_diode_db, i_quench = row[['magnet', 'timestamp_iqps', 'nqps_db', current_col]]
            print('{}/{}: Analysing diode lead resistance of {} on {}, {}'
                  .format(index, index_max, source_qds, Time.to_string(timestamp_qds), timestamp_qds))

            if i_quench < 1000:
                print('The analysis skipped as the quench current is below 1 kA ({} A)'.format(i_quench))
                continue

            # NXCALS
            i_meas_diode_df, u_diode_df = deepcopy(i_meas_u_diode_nxcals_dfs[index])

            if len(u_diode_df) < 1000:
                print('The analysis skipped as the diode measurement has only {} data points'.format(len(u_diode_df)))
                continue

            try:
                res_diode_nxcals_df = DiodeLeadResistanceAnalysis.calculate_resistance_nxcals(circuit_name,
                                                                                              i_meas_diode_df,
                                                                                              u_diode_df)
            except ValueError as err:
                print('The analysis skipped as as we encountered an exception: {}'.format(err))
                continue

            # PM
            if u_diode_db == 'PM':
                i_a_df, u_diode_a_df, u_diode_b_df, u_ref_df = deepcopy(i_meas_u_diode_u_ref_pm_dfs[index])
                if not i_a_df.empty and not u_diode_a_df.empty and not u_diode_b_df.empty and not u_ref_df.empty:
                    try:
                        res_diode_pm_df = \
                            DiodeLeadResistanceAnalysis.calculate_resistance_pm(i_a_df, u_diode_a_df, u_diode_b_df,
                                                                                u_ref_df, timestamp_qds, time_corr=0.0)
                        res_max_pm, t_res_max_pm, _ = \
                            DiodeLeadResistanceAnalysis.extract_res_max_time_current_pm(res_diode_pm_df)
                        DiodeLeadResistancePlot.plot_current_voltage_resistance_pm_nxcals(source_qds, circuit_name,
                                                                                          timestamp_qds,
                                                                                          res_diode_pm_df,
                                                                                          res_diode_nxcals_df)
                    except:
                        res_max_pm = 0
                        t_res_max_pm = 0
                        warnings.warn('Computation of diode resistance failed, '
                                      'please check the input signals and analysis assumptions.', stacklevel=2)


                else:
                    res_max_pm = 0
                    t_res_max_pm = 0
            else:
                res_max_pm = 0
                t_res_max_pm = 0

                DiodeLeadResistancePlot.plot_current_voltage_resistance_nxcals(source_qds, res_diode_nxcals_df)

            # Maximum resistance value
            res_max_cals, t_res_max_cals, _ = DiodeLeadResistanceAnalysis \
                .extract_res_max_time_current_nxcals(res_diode_nxcals_df, 'I_MEAS')

            res_max = max(res_max_pm, res_max_cals)
            t_res_max = max(t_res_max_pm, t_res_max_cals)
            col_r = 'R_DIODE_LEADS_%s' % circuit_name.split('.')[0]
            col_t = 't_DIODE_LEADS_%s' % circuit_name.split('.')[0]

            if res_max > 50:
                warnings.warn('Warning: current leads resistance %s is above 50 uOhm!' % res_max, stacklevel=2)
            elif res_max > 150:
                warnings.warn('Alarm: current leads resistance %s is above 150 uOhm!' % res_max, stacklevel=2)
            else:
                print('Diode lead resistance is %s uOhm.' % res_max)

            # Update results table
            self.results_table.loc[index, col_r] = round(res_max, 2)
            self.results_table.loc[index, col_t] = round(t_res_max, 2)

            # Show next
            check_show_next(index, index_max, self.is_automatic)

    # DFB
    @check_dataframe_empty(mode='all', warning='All DataFrames are empty, LEADS voltage plot skipped!')
    def analyze_leads_voltage(self, u_dfs, circuit_name, timestamp, *, signal, value_min, value_max):

        is_signal_correct = AssertionBuilder().with_signal(u_dfs) \
            .has_min_max_value(value_min=value_min, value_max=value_max) \
            .show_plot(ylabel='{}, [V]'.format(signal),
                       title='%s, %s: %s(t)' % (Time.to_string_short(timestamp), circuit_name, signal)) \
            .get_assertion_result()

        if self.results_table is not None:
            status = '%s < quench threshold' % signal if is_signal_correct else '%s > quench threshold' % signal
            analysis = '%s_LEADS_%s' % (circuit_name, signal)
            self.results_table[analysis] = get_expert_comment(analysis, self.is_automatic, status)

    @staticmethod
    def plot_leads_voltage_zoom(u_hts_dfs, circuit_name, timestamp_fgc, timestamp_ee_odd, timestamp_ee_even,
                                *, signal='U_HTS'):
        title = '%s, %s: %s(t)' % (Time.to_string_short(timestamp_fgc), circuit_name, signal)
        ax = PlotBuilder().with_signal(u_hts_dfs, grid=True, title=title, style='x-') \
            .with_xlim((-1, 2)).with_ylabel(ylabel='%s, [V]' % signal) \
            .plot(show_plot=False) \
            .get_axes()[0]

        ax.axvline((timestamp_ee_odd - timestamp_fgc) * 1e-9, ls='--')
        ax.axvline((timestamp_ee_even - timestamp_fgc) * 1e-9, ls='--', color='green')
        plt.show()

    # VF
    @check_dataframe_empty(mode='all', warning='All DataFrames are empty, nQPS voltage plot skipped!')
    def analyze_voltage_feelers(self, circuit_name, timestamp_fgc, i_meas_df, u_dfs, signal, system,
                                xlim=None, ylim=None):
        # Plot
        title = '%s %s I_MEAS(t) and %s(t)' % (Time.to_string_short(timestamp_fgc), circuit_name, signal)
        PlotBuilder().with_signal(i_meas_df, title=title, grid=True) \
            .with_ylabel(ylabel='I_MEAS, [A]') \
            .with_signal(u_dfs, legend=False) \
            .with_ylabel(ylabel='%s, [V]' % signal) \
            .with_xlim(xlim).with_ylim(ylim) \
            .plot()

        PlotBuilder().with_signal(i_meas_df, title=title, grid=True) \
            .with_ylabel(ylabel='I_MEAS, [A]') \
            .with_signal(u_dfs, legend=False) \
            .with_ylabel(ylabel='%s, [V]' % signal) \
            .with_xlim((-2, 3)).with_ylim(ylim) \
            .plot()

        # Analyze
        not_communicating_cards = []
        disabled_cards = []
        status = 'OK'
        for u_df in u_dfs:
            if (float(u_df.min()) < -2000) and (float(u_df.max()) < -2000):
                not_communicating_cards.append(u_df.columns[0])
                status = 'NOT_OK'
            if (float(u_df.min()) > -1) and (float(u_df.max()) < 1):
                disabled_cards.append(u_df.columns[0])
                status = 'NOT_OK'
        if not_communicating_cards:
            status += '_NOT_COMMUNICATING_CARDS'
        if disabled_cards:
            status += '_DISABLED_CARDS'

        # Print a report
        print('Report on Voltage Feeler cards:')
        print('- there are {} cards.'.format(len(u_dfs)))
        print('- there are {} functioning cards.'
              .format(len(u_dfs) - len(not_communicating_cards) - len(disabled_cards)))
        print('- there are {} not communicating cards: {}.'
              .format(len(not_communicating_cards), not_communicating_cards))
        print('- there are {} disabled cards: {}.'.format(len(disabled_cards), disabled_cards))

        # Store analysis result
        if self.results_table is not None:
            system_signal = '{}_{}'.format(system, signal)
            self.results_table[system_signal] = get_expert_comment(system_signal, self.is_automatic, status)

    def calculate_resistance(self, i_meas_feature_df, u_res_feature_df, signal_name, t_start, circuit_name):
        return ResistanceBuilder().with_busbar_voltage(u_res_feature_df) \
            .with_busbar_current(i_meas_feature_df) \
            .calculate_mean_resistance(self.circuit_type) \
            .convert_to_row(index=t_start) \
            .fill_missing_columns(signal_name=signal_name.replace('U', 'R'), circuit_type=self.circuit_type) \
            .sort_columns_busbar_location(signal_name=signal_name.replace('U', 'R'), circuit_type=self.circuit_type,
                                          circuit_name=circuit_name, split='_') \
            .df

    @staticmethod
    def analyze_busbar_magnet_resistance(res_df, signal_name, max_value):
        return AssertionBuilder().with_feature(res_df) \
            .has_min_max_value(feature=signal_name, min_value=0, max_value=max_value) \
            .show_plot(xlabel='Busbar, [-]', ylabel=('%s (Calculated), [Ohm]' % signal_name)) \
            .get_features_outside_range()

    # DFB
    @staticmethod
    def assert_tt891a_min_max_value(tt891a_dfs, value_range=(46, 54)):
        return AssertionBuilder().with_signal(tt891a_dfs) \
                .has_min_max_value(value_min=value_range[0], value_max=value_range[1]) \
                .show_plot(ylabel='TT891A.TEMPERATURECALC [K]')

    @staticmethod
    def assert_tt893_min_max_value(tt893_dfs, value_range=(280, 320)):
        return AssertionBuilder().with_signal(tt893_dfs) \
            .has_min_max_value(value_min=value_range[0], value_max=value_range[1]) \
            .show_plot(ylabel='TT893.TEMPERATURECALC, [K]')

    @staticmethod
    def assert_u_res_min_max_value(u_res_dfs, value_range=(-40e-3, 40e-3)):
        return AssertionBuilder().with_signal(u_res_dfs) \
            .has_min_max_value(value_min=value_range[0], value_max=value_range[1]) \
            .show_plot(ylabel='U_RES, [V]')

    @staticmethod
    def assert_u_res_min_max_slope(u_res_dfs, plateau_start, plateau_end, t_0, slope_range=(-2, 2)):
        return AssertionBuilder().with_signal(u_res_dfs) \
            .with_time_range(t_start=(np.array(plateau_start) - t_0) / 1e9,
                             t_end=(np.array(plateau_end) - t_0) / 1e9) \
            .has_min_max_slope(slope_min=slope_range[0], slope_max=slope_range[1]) \
            .show_plot(ylabel='U_RES, [V]')

    @staticmethod
    def assert_u_hts_min_max_value(u_hts_dfs, value_range=(-0.5e-3, 0.5e-3)):
        return AssertionBuilder().with_signal(u_hts_dfs) \
            .has_min_max_value(value_min=value_range[0], value_max=value_range[1]) \
            .show_plot(ylabel='U_HTS, [V]')

    @staticmethod
    def assert_u_hts_min_max_slope(u_hts_dfs, plateau_start, plateau_end, t_0, slope_range=(-0.5, 0.5)):
        return AssertionBuilder().with_signal(u_hts_dfs) \
                .with_time_range(t_start=(np.array(plateau_start) - t_0) / 1e9,
                                 t_end=(np.array(plateau_end) - t_0) / 1e9) \
                .has_min_max_slope(slope_min=slope_range[0], slope_max=slope_range[1]) \
                .show_plot(ylabel='U_HTS, [V]')

    @staticmethod
    def assert_u_res_between_two_references(u_res_dfs, first_ref, first_scaling, second_ref, second_scaling):
        return AssertionBuilder().with_signal(u_res_dfs) \
            .is_between_two_references(first_ref=first_ref, first_scaling=first_scaling,
                                       second_ref=second_ref, second_scaling=second_scaling) \
            .show_plot(ylabel='U_RES, [V]', data_sec=first_ref, ylabel_sec='I_MEAS, [A]')

    @staticmethod
    def assert_cv891_min_max_variation(cv891_dfs, variation_min_max, plateau_start, plateau_end, t_0):
        return AssertionBuilder().with_signal(cv891_dfs) \
            .with_time_range(t_start=(np.array(plateau_start)-t_0)/1e9,
                             t_end=(np.array(plateau_end)-t_0)/1e9) \
            .has_min_max_variation(variation_min_max=variation_min_max) \
            .show_plot(ylabel='CV891.POSST. [%]')

    @staticmethod
    def assert_cv891_min_max_value(cv891_dfs, value_range=(2, 50)):
        return AssertionBuilder().with_signal(cv891_dfs) \
            .has_min_max_value(value_min=value_range[0], value_max=value_range[1]) \
            .show_plot(ylabel='CV891.POSST, [%]')

