from functools import wraps
from typing import Union, List, Tuple
from IPython.display import display, HTML
import matplotlib.pyplot as plt
from pyspark.sql.types import IntegerType
import pyspark.sql.functions as F

import pandas as pd
import numpy as np
from tqdm.notebook import tqdm
import warnings

from lhcsmapi.analysis.busbar.BusbarResistanceAnalysis import find_start_end_of_the_longest_ramp_up, \
    find_plateau_start_and_end, extract_intersecting_plateaus
from lhcsmapi.analysis.decorators import check_nan_timestamp, check_nan_timestamp_signals
from lhcsmapi.analysis.expert_input import check_show_next
from lhcsmapi.pyedsl.QueryBuilder import QueryBuilder
from lhcsmapi.reference.Reference import Reference
from lhcsmapi.metadata.MappingMetadata import MappingMetadata
from lhcsmapi.Time import Time


def execution_count():
    def inner_function(func):
        @wraps(func)
        def wrapper(*args, **kw):
            query = list(filter(lambda arg: isinstance(arg, CircuitQuery), args))[0]
            if query.max_executions is not None:
                query.execution_count += 1

                def cstr(s, color='black'):
                    return "<text style=color:{}>{}</text>".format(color, s)

                output_text = 'Executing %s query function %s: %d/%d.' \
                              % (query.circuit_name, func.__name__, query.execution_count, query.max_executions)
                display(HTML(cstr(output_text, 'blue')))
            return func(*args, **kw)

        return wrapper

    return inner_function


def get_quench_current(i_meas_df, index):
    value_at_index = i_meas_df.reindex([index], method='nearest', tolerance=.1).values[0]
    if np.isnan(value_at_index):
        value_at_index = i_meas_df.values[-1]
        warnings.warn('Time of quench {} s is outside of PM FGC I_MEAS signal. Taken last value of current. '
                      'For a more accurate value of quench current, please query NXCALS.'.format(index))
    return int(value_at_index)


class CircuitQuery(object):
    """Class containing methods for query of events and signals. Each query is implemented with pyeDSL.

    """

    def __init__(self, circuit_type, circuit_name, max_executions=None, verbose=True):
        self.circuit_type = circuit_type
        self.circuit_name = circuit_name
        self.verbose = verbose
        self.max_executions = max_executions
        self.execution_count = 0

    # General purpose
    @execution_count()
    def query_signal_nxcals(self, t_start, t_end=None, duration=None, *, system, signal_names, spark):
        t_start, t_end = Time.get_query_period_in_unix_time(t_start, t_end, duration, '')
        return QueryBuilder().with_nxcals(spark) \
            .with_duration(t_start=t_start, t_end=t_end) \
            .with_circuit_type(self.circuit_type) \
            .with_metadata(circuit_name=self.circuit_name, system=system, signal=signal_names) \
            .signal_query() \
            .synchronize_time() \
            .convert_index_to_sec().dfs

    @execution_count()
    def query_raw_signal_nxcals(self, t_start, t_end=None, duration=None, *, system, signal_names, spark):
        t_start, t_end = Time.get_query_period_in_unix_time(t_start, t_end, duration, '')
        return QueryBuilder().with_nxcals(spark) \
            .with_duration(t_start=t_start, t_end=t_end) \
            .with_circuit_type(self.circuit_type) \
            .with_metadata(circuit_name=self.circuit_name, system=system, signal=signal_names) \
            .signal_query().dfs

    # PIC
    @execution_count()
    @check_nan_timestamp(return_type=np.nan)
    def find_timestamp_pic(self, timestamp_fgc: int, *, spark, duration=[(1, 's'), (60, 's')]) -> float:
        """ Method searching PIC timestamps around an FGC PM event timestamp.
        If the input timestamp is NaN or the PIC timestamp was not found, then a NaN is returned.

        :param timestamp_fgc: FGC PM event timestamp (ns precision)
        :type timestamp_fgc: int
        :param spark: spark session variable needed to perform an NXCALS query
        :type spark: spark session
        :param duration: search duration before and after the FGC PM timestamp
        :type duration: List[Tuple[int, str]]
        :return: float value of PIC timestamp in ns precision
        """
        st_abort_pic_df = QueryBuilder().with_nxcals(spark) \
            .with_duration(t_start=timestamp_fgc, duration=duration) \
            .with_circuit_type(self.circuit_type) \
            .with_metadata(circuit_name=self.circuit_name, system='PIC', signal='ST_ABORT_PIC') \
            .signal_query(verbose=self.verbose).dfs[0]

        return st_abort_pic_df.index[0] if len(st_abort_pic_df) > 0 else np.nan

    # PC
    @execution_count()
    @check_nan_timestamp(return_type=pd.DataFrame())
    def find_source_timestamp_pc(self, t_start: Union[str, int, float], t_end: Union[str, int, float]) -> pd.DataFrame:
        """ Method searching a PM FGC events between given start and end times.
        If either the start or end time is NaN, then an empty dataframe is returned.

        :param t_start: start time of a search
        :type t_start: Union[str, int, float]
        :param t_end: end time of a search
        :type t_end: Union[str, int, float]
        :return: pd.DataFrame with source and timestamp of the PM FGC events
        """
        return QueryBuilder().with_pm() \
            .with_duration(t_start=t_start, t_end=t_end) \
            .with_circuit_type(self.circuit_type) \
            .with_metadata(circuit_name=self.circuit_name, system='PC') \
            .event_query(verbose=self.verbose).df

    @execution_count()
    @check_nan_timestamp_signals(return_type=pd.DataFrame())
    def query_pc_nxcals(self, timestamp_fgc: int, *, signal_names: List[str], spark, duration=[(300, 's'), (10, 's')],
                        ) -> List[pd.DataFrame]:
        """ Method querying NXCALS for iQPS signals

        :param timestamp_fgc: FGC PM event timestamp (ns precision) to which the query is synchronised
        :type timestamp_fgc: int
        :param signal_names: list of signal names to query according to metadata description
        :type signal_names: List[str]
        :param spark: spark session variable needed to perform an NXCALS query
        :type spark: spark session
        :param duration: search duration before and after the start time
        :type duration: List[Tuple[int, str]]
        :return: list of pd.DataFrames (if a signal was not queried, an empty pd.DataFrame is returned and warning
            containing the raw query is returned)
        """
        return QueryBuilder().with_nxcals(spark) \
            .with_duration(t_start=timestamp_fgc, duration=duration) \
            .with_circuit_type(self.circuit_type) \
            .with_metadata(circuit_name=self.circuit_name, system='PC', signal=signal_names) \
            .signal_query(verbose=self.verbose) \
            .synchronize_time(timestamp_fgc) \
            .convert_index_to_sec() \
            .filter_median().dfs

    @execution_count()
    @check_nan_timestamp_signals(return_type=pd.DataFrame())
    def query_pc_pm(self, timestamp_fgc: int, timestamp_sync: int, *, signal_names: List[str]) -> List[pd.DataFrame]:
        """ Method querying PM for power converter signals
        If any timestamp is NaN, then a list of empty dataframes of the size of the signal names is returned.
        The signals are synchronized in time and converted from ns to s.

        :param timestamp_fgc: FGC PM event timestamp (ns precision)
        :type timestamp_fgc: int
        :param timestamp_sync: timestamp to which the queried signals are synchronized (ns precision)
        :type timestamp_sync: int
        :param signal_names: list of signal names to query according to metadata description
        :type signal_names: List[str]
        :return: list of pd.DataFrames (if a signal was not queried, an empty pd.DataFrame is returned and warning
            containing the raw query is returned)
        """
        return QueryBuilder().with_pm() \
            .with_timestamp(timestamp_fgc) \
            .with_circuit_type(self.circuit_type) \
            .with_metadata(circuit_name=self.circuit_name, system='PC', signal=signal_names) \
            .signal_query(verbose=self.verbose) \
            .synchronize_time(timestamp_sync) \
            .convert_index_to_sec().dfs

    @execution_count()
    @check_nan_timestamp_signals(return_type=pd.DataFrame())
    def query_pc_pm_events(self, timestamp_fgc: int, *, signal_names: List[str]) -> None:
        """ Method querying PM for FGC event signals
        If the timestamp is NaN, then a list of empty dataframes of the size of the signal names is returned.
        Note that the output signal may have duplicated indices and is neither synchronized nor converted to s.

        :param timestamp_fgc: FGC PM event timestamp (ns precision)
        :type timestamp_fgc: int
        :param signal_names: list of signal names to query according to metadata description
        :type signal_names: List[str]
        :return: list of pd.DataFrames (if a signal was not queried, an empty pd.DataFrame is returned and warning
            containing the raw query is returned)
        """
        return QueryBuilder().with_pm() \
            .with_timestamp(timestamp_fgc) \
            .with_circuit_type(self.circuit_type) \
            .with_metadata(circuit_name=self.circuit_name, system='PC', signal=signal_names) \
            .signal_query(allow_duplicate_index=True, verbose=self.verbose).dfs

    @execution_count()
    def get_timestamp_ref(self, col='fgcPm') -> int:
        """ Method returning a reference timestamp for either FGC or EE

        :param col: name of a column with information about reference timestamp (either fgcPm or eePm)
        :return: unix timestamp in ns
        """
        timestamp_ref = Reference.get_power_converter_reference_fpa(self.circuit_type, self.circuit_name, col)
        return Time.to_unix_timestamp(timestamp_ref)

    # EE
    @execution_count()
    @check_nan_timestamp(return_type=pd.DataFrame())
    def find_source_timestamp_ee(self, t_start: Union[str, int, float], duration=[(10, 's'), (1000, 's')],
                                 system='EE') -> pd.DataFrame:
        """ Method searching PM energy extraction events
        If either the start or end time is NaN, then an empty dataframe is returned.

        :param t_start: start time of a query
        :param duration: search duration before and after the start time
        :type duration: List[Tuple[int, str]]
        :param system: name of an EE system ('EE' by default, however for RB it takes either 'EE_ODD' or 'EE_EVEN')
        :type system: str
        return: pd.DataFrame with source and timestamp of the PM EE events
        """
        return QueryBuilder().with_pm() \
            .with_duration(t_start=t_start, duration=duration) \
            .with_circuit_type(self.circuit_type) \
            .with_metadata(circuit_name=self.circuit_name, system=system) \
            .event_query(verbose=self.verbose).df

    @execution_count()
    @check_nan_timestamp_signals(return_type=pd.DataFrame())
    def query_ee_u_dump_res_pm(self, timestamp_ee: int, timestamp_sync: int, system='EE', *,
                               signal_names: List[str]) -> List[pd.DataFrame]:
        """ Method querying PM for energy extraction voltage signals
        If the timestamp is NaN, then a list of empty dataframes of the size of the signal names is returned.
        The signals are synchronized in time and converted from ns to s.

        :param timestamp_ee: EE PM event timestamp (ns precision)
        :type timestamp_ee: int
        :param timestamp_sync: timestamp to which the queried signals are synchronized (ns precision)
        :type timestamp_sync: int
        :param system: name of EE system to query
        :type system: str
        :param signal_names: list of signal names to query according to metadata description
        :type signal_names: List[str]
        :return: list of pd.DataFrames (if a signal was not queried, an empty pd.DataFrame is returned and warning
            containing the raw query is returned)
        """
        return QueryBuilder().with_pm() \
            .with_timestamp(timestamp_ee) \
            .with_circuit_type(self.circuit_type) \
            .with_metadata(circuit_name=self.circuit_name, system=system, signal=signal_names) \
            .signal_query(verbose=self.verbose) \
            .remove_initial_offset(timestamp_ee) \
            .synchronize_time(timestamp_sync) \
            .convert_index_to_sec().dfs

    @execution_count()
    @check_nan_timestamp_signals(return_type=pd.DataFrame())
    def query_ee_t_res_pm(self, timestamp_ee: int, timestamp_sync: int, system='EE', *, signal_names: List[str]) \
            -> List[pd.DataFrame]:
        """ Method querying PM for energy extraction temperature signals
        If the timestamp is NaN, then a list of empty dataframes of the size of the signal names is returned.
        The signals are synchronized in time and converted from ns to s.

        :param timestamp_ee: EE PM event timestamp (ns precision)
        :type timestamp_ee: int
        :param timestamp_sync: timestamp to which the queried signals are synchronized (ns precision)
        :type timestamp_sync: int
        :param system: name of EE system to query
        :type system: str
        :param signal_names: list of signal names to query according to metadata description
        :type signal_names: List[str]
        :return: list of pd.DataFrames (if a signal was not queried, an empty pd.DataFrame is returned and warning
            containing the raw query is returned)
        """
        return QueryBuilder().with_pm() \
            .with_timestamp(timestamp_ee) \
            .with_circuit_type(self.circuit_type) \
            .with_metadata(circuit_name=self.circuit_name, system=system, signal=signal_names) \
            .signal_query(verbose=self.verbose) \
            .remove_values_for_time_less_than(timestamp_ee) \
            .synchronize_time(timestamp_sync) \
            .convert_index_to_sec().dfs

    @execution_count()
    @check_nan_timestamp_signals(return_type=pd.DataFrame())
    def query_ee_cals(self, timestamp_fgc: Union[int, str, float], system='EE', *, signal_names: List[str], ldb,
                      duration=[(3 * 3600, 's')], t_thr=1000) -> List[pd.DataFrame]:
        """ Method querying CALS for energy extraction signals
        If the timestamp is NaN, then a list of empty dataframes of the size of the signal names is returned.
        The signals are synchronized in time and converted from ns to s.

        :param timestamp_fgc: FGC PM event timestamp (ns precision) to which the query is synchronised
        :type timestamp_fgc: int
        :param system: name of EE system
        :type system: str
        :param signal_names: list of signal names to query according to metadata description
        :type signal_names: List[str]
        :param ldb: CALS pytimber connector
        :type ldb: pytimber
        :param duration: search duration before and after the start time
        :type duration: List[Tuple[int, str]]
        :param t_thr: threshold of time for the query after which signals are taken for the output
        :type t_thr: int
        :return: list of pd.DataFrames (if a signal was not queried, an empty pd.DataFrame is returned and warning
            containing the raw query is returned)
        """
        return QueryBuilder().with_cals(ldb) \
            .with_duration(t_start=timestamp_fgc, duration=duration) \
            .with_circuit_type(self.circuit_type) \
            .with_metadata(circuit_name=self.circuit_name, system=system, signal=signal_names) \
            .signal_query(verbose=self.verbose) \
            .synchronize_time(timestamp_fgc / 1e9) \
            .remove_values_for_time_less_than(t_thr).dfs

    # iQPS
    @execution_count()
    @check_nan_timestamp_signals(return_type=pd.DataFrame())
    def query_iqps_nxcals(self, timestamp_fgc: int, *, signal_names: List[str], spark, duration=[(300, 's'), (10, 's')],
                          ) -> List[pd.DataFrame]:
        """ Method querying NXCALS for iQPS signals

        :param timestamp_fgc: FGC PM event timestamp (ns precision) to which the query is synchronised
        :type timestamp_fgc: int
        :param signal_names: list of signal names to query according to metadata description
        :type signal_names: List[str]
        :param spark: spark session variable needed to perform an NXCALS query
        :type spark: spark session
        :param duration: search duration before and after the start time
        :type duration: List[Tuple[int, str]]
        :return: list of pd.DataFrames (if a signal was not queried, an empty pd.DataFrame is returned and warning
            containing the raw query is returned)
        """
        return QueryBuilder().with_nxcals(spark) \
            .with_duration(t_start=timestamp_fgc, duration=duration) \
            .with_circuit_type(self.circuit_type) \
            .with_metadata(circuit_name=self.circuit_name, system='QDS', signal=signal_names) \
            .signal_query(verbose=self.verbose) \
            .synchronize_time(timestamp_fgc) \
            .convert_index_to_sec() \
            .filter_median().dfs

    @execution_count()
    @check_nan_timestamp(return_type=pd.DataFrame())
    def find_source_timestamp_qds(self, t_start: Union[int, str, float], duration=[(10, 's'), (400, 's')],
                                  ) -> pd.DataFrame:
        """ Method searching PM QDS events
        If the start is NaN, then an empty dataframe is returned.

        :param t_start: start time of a search
        :type t_start: Union[str, int, float]
        :param duration: search duration before and after the start time
        :type duration: List[Tuple[int, str]]
        :return: pd.DataFrame with source and timestamp of the PM QDS events
        """
        return QueryBuilder().with_pm() \
            .with_duration(t_start=t_start, duration=duration) \
            .with_circuit_type(self.circuit_type) \
            .with_metadata(circuit_name=self.circuit_name, system='QDS', source='*') \
            .event_query(verbose=self.verbose) \
            .filter_source(self.circuit_name, 'QDS') \
            .drop_duplicate_source() \
            .sort_values(by='timestamp').df

    @execution_count()
    @check_nan_timestamp_signals(return_type=pd.DataFrame())
    def query_qds_pm(self, timestamp_qds: int, timestamp_qds_sync: int, *, signal_names: List[str]
                     ) -> List[pd.DataFrame]:
        """ Method querying PM for QDS signals
        If the timestamp is NaN, then a list of empty dataframes of the size of the signal names is returned.
        The signals are synchronized in time and converted from ns to s.

        :param timestamp_qds: PM QDS timestamp in ns precision
        :type timestamp_qds: int
        :param timestamp_qds: PM QDS timestamp in ns precision (for signal synchronization)
        :type timestamp_qds: int
        :param signal_names: list of signal names to query according to metadata description
        :type signal_names: List[str]
        :return: list of pd.DataFrames (if a signal was not queried, an empty pd.DataFrame is returned and warning
            containing the raw query is returned)
        """
        return QueryBuilder().with_pm() \
            .with_timestamp(timestamp_qds) \
            .with_circuit_type(self.circuit_type) \
            .with_metadata(circuit_name=self.circuit_name, system='QDS', signal=signal_names) \
            .signal_query(verbose=self.verbose) \
            .synchronize_time(timestamp_qds_sync) \
            .convert_index_to_sec() \
            .filter_median().dfs

    @execution_count()
    def query_iqps_analog_pm(self, source_timestamp_qds_df: pd.DataFrame,
                             signal_names: List[str]) -> List[List[pd.DataFrame]]:
        """ Method querying PM for iQPS analog signals
        Signals are synchronized to the iQPS PM event and index is converted to seconds.

        :param source_timestamp_qds_df: dataframe with PM source and timestamp for QDS events
        :type source_timestamp_qds_df: pd.DataFrame
        :param signal_names: list of signal names to query according to metadata description
        :type signal_names: List[str]
        :return: List of list of iQPS signals. The outer list has as many elements as iQPS events and the lenght
            of the inner is equal to the number of iQPS signals.
        """
        iqps_analog_dfs = []
        for index, row in tqdm(source_timestamp_qds_df.iterrows(), total=source_timestamp_qds_df.shape[0],
                               desc='Querying iQPS PM'):
            timestamp_qds = row['timestamp']
            source_qds = row['source']

            iqps_analog_df = QueryBuilder().with_pm() \
                .with_timestamp(timestamp_qds) \
                .with_circuit_type(self.circuit_type) \
                .with_metadata(circuit_name=self.circuit_name, system='QDS', signal=signal_names, source=source_qds,
                               wildcard={'CELL': source_qds}) \
                .signal_query() \
                .synchronize_time(timestamp_qds) \
                .convert_index_to_sec() \
                .filter_median().dfs
            iqps_analog_dfs.append(iqps_analog_df)

        return iqps_analog_dfs

    @execution_count()
    def query_iqps_digital_pm(self, source_timestamp_qds_df: pd.DataFrame,
                              signal_names: List[str]) -> List[List[pd.DataFrame]]:
        """ Method querying PM for iQPS digital signals
        Signals are synchronized to the iQPS PM event and index is converted to seconds.

        :param source_timestamp_qds_df: dataframe with PM source and timestamp for QDS events
        :type source_timestamp_qds_df: pd.DataFrame
        :param signal_names: list of signal names to query according to metadata description
        :type signal_names: List[str]
        :return: List of list of iQPS signals. The outer list has as many elements as iQPS events and the lenght
            of the inner is equal to the number of iQPS signals.
        """
        iqps_digital_dfs = []
        for index, row in tqdm(source_timestamp_qds_df.iterrows(), total=source_timestamp_qds_df.shape[0],
                               desc='Querying IQPS PM'):
            timestamp_qds = row['timestamp']
            source_qds = row['source']

            iqps_digital_df = QueryBuilder().with_pm() \
                .with_timestamp(timestamp_qds) \
                .with_circuit_type(self.circuit_type) \
                .with_metadata(circuit_name=self.circuit_name, system='QDS', signal=signal_names, source=source_qds) \
                .signal_query() \
                .synchronize_time(timestamp_qds) \
                .convert_index_to_sec().dfs
            iqps_digital_dfs.append(iqps_digital_df)

        return iqps_digital_dfs

    # nQPS
    @execution_count()
    def find_source_timestamp_nqps(self, t_start: Union[int, str, float], duration=[(10, 's'), (400, 's')],
                                   ) -> pd.DataFrame:
        """ Method searching PM nQPS events
        If the start is NaN, then an empty dataframe is returned.

        :param t_start: start time of a search
        :type t_start: Union[str, int, float]
        :param duration: search duration before and after the start time
        :type duration: List[Tuple[int, str]]
        :return: pd.DataFrame with source and timestamp of the PM nQPS events
        """
        system = 'DIODE_{}'.format(self.circuit_name.split('.')[0])
        return QueryBuilder().with_pm() \
            .with_duration(t_start=t_start, duration=duration) \
            .with_circuit_type(self.circuit_type) \
            .with_metadata(circuit_name=self.circuit_name, system=system, source='*') \
            .event_query(verbose=self.verbose) \
            .filter_source(self.circuit_name, 'QDS') \
            .drop_duplicate_source() \
            .sort_values(by='timestamp').df

    @execution_count()
    def query_nqps_voltage_pm(self, source_timestamp_qds_df: pd.DataFrame) -> List[List[pd.DataFrame]]:
        """ Method querying PM for nQPS signals
        Signals are synchronized to the nQPS PM event and index is converted to seconds.

        :param source_timestamp_qds_df: dataframe with PM source and timestamp for QDS events
        :type source_timestamp_qds_df: pd.DataFrame
        :return: List of list of nQPS signals. The outer list has as many elements as nQPS events and the lenght
            of the inner is equal to the number of nQPS signals.
        """
        u_nqps_dfs = []
        for index, row in tqdm(source_timestamp_qds_df.iterrows(), total=source_timestamp_qds_df.shape[0],
                               desc='Querying PM for nQPS analog signals'):
            timestamp_qds = row['timestamp']
            source_qds = row['source']
            qps_crate = MappingMetadata.get_crate_name_from_magnet_name(self.circuit_type, source_qds)
            system = 'DIODE_{}'.format(self.circuit_name.split('.')[0])
            signal = 'U_DIODE_{}'.format(self.circuit_name.split('.')[0])
            wildcard = 'MB.{}'.format(source_qds) if self.circuit_type == 'RB' else 'MQ.{}'.format(source_qds)

            source_timestamp_nqps_df = self.find_source_timestamp_nqps_pm(source_qds, timestamp_qds)
            if not source_timestamp_nqps_df.empty:
                timestamp_nqps = source_timestamp_nqps_df.loc[0, 'timestamp']

                u_diode_rqd_df = QueryBuilder().with_pm() \
                    .with_timestamp(timestamp_nqps) \
                    .with_circuit_type(self.circuit_type) \
                    .with_metadata(circuit_name=self.circuit_name, system=system, signal=signal, source=qps_crate,
                                   wildcard={'MAGNET': wildcard}) \
                    .signal_query() \
                    .synchronize_time(timestamp_nqps) \
                    .convert_index_to_sec().dfs[0]

                u_ref_rqd_df = QueryBuilder().with_pm() \
                    .with_timestamp(timestamp_nqps) \
                    .with_circuit_type(self.circuit_type) \
                    .with_metadata(circuit_name=self.circuit_name, system=system, signal='U_REF_N1', source=qps_crate,
                                   wildcard={'QPS_CRATE': qps_crate}) \
                    .signal_query() \
                    .synchronize_time(timestamp_nqps) \
                    .convert_index_to_sec().dfs[0]
                u_nqps_dfs.append([u_diode_rqd_df, u_ref_rqd_df])
            else:
                warnings.warn('nQPS buffer is empty for %s !' % source_qds)
                u_nqps_dfs.append([pd.DataFrame(), pd.DataFrame()])

        return u_nqps_dfs

    # QH
    @execution_count()
    def query_qh_pm(self, source_timestamp_qds_df: pd.DataFrame, *, signal_names: List[str],
                    is_ref=False) -> List[List[pd.DataFrame]]:
        """ Method querying PM for QH signals (either voltage or current or both)
        Signals are synchronized to the QH PM event and index is converted to seconds.

        :param source_timestamp_qds_df: dataframe with PM source and timestamp for QDS events
        :type source_timestamp_qds_df: pd.DataFrame
        :param signal_names: list of signal names to query according to metadata description
        :type signal_names: List[str]
        :param is_ref: boolean flag to determine whether a query of reference or actual discharge is necessary
        :type is_ref: bool
        :return: List of list of QH signals. The outer list has as many elements as QH events and the lenght
            of the inner is equal to the number of QH signals.
        """
        hds_dfs = []
        for index, row in tqdm(source_timestamp_qds_df.iterrows(), total=source_timestamp_qds_df.shape[0],
                               desc='Querying PM'):
            source_qds = row['source']
            if is_ref:
                timestamp_qh = Reference.get_quench_heater_reference_discharge(self.circuit_type, source_qds)
            else:
                if self.circuit_type in ['RB', 'RQ']:
                    timestamp_qh = row['timestamp'] + 1000000
                else:
                    timestamp_qh = row['timestamp']

            hds_df = QueryBuilder().with_pm() \
                .with_timestamp(timestamp_qh) \
                .with_circuit_type(self.circuit_type) \
                .with_metadata(circuit_name=self.circuit_name, system='QH', signal=signal_names, source=source_qds,
                               wildcard={'CELL': source_qds}) \
                .signal_query() \
                .synchronize_time(timestamp_qh) \
                .convert_index_to_sec().dfs

            hds_dfs.append(hds_df)

        return hds_dfs

    # DIODE LEADS
    @execution_count()
    def find_source_timestamp_nqps_pm(self, source_qds: str, timestamp_qds: int, duration=[(600, 's'), (600, 's')],
                                      ) -> pd.DataFrame:
        """ Method searching PM nQPS events.
        If the timestamp is NaN, then an empty dataframe is returned.

        :param source_qds: PM QDS source according to metadata
        :type source_qds: str
        :param timestamp_qds: PM QDS timestamp in ns precision
        :type timestamp_qds: int
        :param duration: search duration before and after the start time
        :type duration: List[Tuple[int, str]]
        :return: pd.DataFrame with source and timestamp of the PM nQPS events
        """
        qps_crate = MappingMetadata.get_crate_name_from_magnet_name(self.circuit_type, source_qds)
        system = 'DIODE_{}'.format(self.circuit_name.split('.')[0])

        return QueryBuilder().with_pm() \
            .with_duration(t_start=timestamp_qds, duration=duration) \
            .with_circuit_type(self.circuit_type) \
            .with_metadata(circuit_name=self.circuit_name, system=system, source=qps_crate) \
            .event_query(verbose=self.verbose).df

    def _query_i_u_diode_pm(self, circuit_name: str, source_qds: str, timestamp_fgc: int, timestamp_qds_a: int,
                            timestamp_qds_b: int) -> Tuple[pd.DataFrame, pd.DataFrame, pd.DataFrame, pd.DataFrame]:
        """ Method querying PM for current and voltage of a diode
        Signals are **not** synchronized to the input timestamp and index is **not** converted to seconds.

        :param circuit_name: name of a circuit to query
        :type circuit_name: str
        :param source_qds: PM QDS source according to metadata
        :type source_qds: str
        :param timestamp_fgc: FGC PM event timestamp for I_A query (ns precision)
        :type timestamp_fgc: int
        :param timestamp_qds_a: QDS Board A PM event timestamp (ns precision)
        :type timestamp_qds_a: int
        :param timestamp_qds_b: QDS Board B PM event timestamp (ns precision)
        :type timestamp_qds_b: int
        :return: tuple of pd.DataFrames (if a signal was not queried, an empty pd.DataFrame is returned and warning
            containing the raw query is returned)
        """
        i_a_rqd_df: pd.DataFrame = QueryBuilder().with_pm() \
            .with_timestamp(timestamp_fgc) \
            .with_circuit_type(self.circuit_type) \
            .with_metadata(circuit_name=circuit_name, system='PC', signal='I_A') \
            .signal_query().dfs[0]

        qps_crate = MappingMetadata.get_crate_name_from_magnet_name(self.circuit_type, source_qds)
        system = 'DIODE_{}'.format(circuit_name.split('.')[0])
        signal = 'U_DIODE_{}'.format(circuit_name.split('.')[0])
        wildcard = 'MB.{}'.format(source_qds) if self.circuit_type == 'RB' else 'MQ.{}'.format(source_qds)

        u_diode_a_rqd_df: pd.DataFrame = QueryBuilder().with_pm() \
            .with_timestamp(timestamp_qds_a) \
            .with_circuit_type(self.circuit_type) \
            .with_metadata(circuit_name=circuit_name, system=system, signal=signal, source=qps_crate,
                           wildcard={'MAGNET': wildcard}) \
            .signal_query().dfs[0]

        u_diode_b_rqd_df: pd.DataFrame = QueryBuilder().with_pm() \
            .with_timestamp(timestamp_qds_b) \
            .with_circuit_type(self.circuit_type) \
            .with_metadata(circuit_name=circuit_name, system=system, signal=signal, source=qps_crate,
                           wildcard={'MAGNET': wildcard}) \
            .signal_query().dfs[0]

        u_ref_rqd_df: pd.DataFrame = QueryBuilder().with_pm() \
            .with_timestamp(timestamp_qds_a) \
            .with_circuit_type(self.circuit_type) \
            .with_metadata(circuit_name=circuit_name, system=system, signal='U_REF_N1', source=qps_crate,
                           wildcard={'QPS_CRATE': qps_crate}) \
            .signal_query().dfs[0]

        return i_a_rqd_df, u_diode_a_rqd_df, u_diode_b_rqd_df, u_ref_rqd_df

    def _query_i_u_diode_nxcals(self, circuit_name: str, source_qds: str, timestamp_qds: int, *, spark,
                                duration) -> Tuple[pd.DataFrame, pd.DataFrame]:
        """ Method querying NXCALS for current and voltage of a diode
        Signals are synchronized to the input timestamp and index is converted to seconds.

        :param circuit_name: name of a circuit to query
        :type circuit_name: str
        :param source_qds: PM QDS source according to metadata
        :type source_qds: str
        :param timestamp_qds: QDS Board A PM event timestamp (ns precision) to which the query is synchronised
        :type timestamp_qds: int
        :param spark: spark session variable needed to perform an NXCALS query
        :type spark: spark session
        :param duration: search duration before and after the start time
        :type duration: List[Tuple[int, str]]
        :return: tuple of pd.DataFrames (if a signal was not queried, an empty pd.DataFrame is returned and warning
            containing the raw query is returned)
        """
        i_meas_diode_df: pd.DataFrame = QueryBuilder().with_nxcals(spark) \
            .with_duration(t_start=timestamp_qds, duration=duration) \
            .with_circuit_type(self.circuit_type) \
            .with_metadata(circuit_name=circuit_name, system='PC', signal='I_MEAS') \
            .signal_query() \
            .synchronize_time(timestamp_qds) \
            .convert_index_to_sec().dfs[0]

        system = 'DIODE_{}'.format(circuit_name.split('.')[0])
        signal = 'U_DIODE_{}'.format(circuit_name.split('.')[0])
        wildcard = 'MB.{}'.format(source_qds) if self.circuit_type == 'RB' else 'MQ.{}'.format(source_qds)

        u_diode_rqd_df: pd.DataFrame = QueryBuilder().with_nxcals(spark) \
            .with_duration(t_start=timestamp_qds, duration=duration) \
            .with_circuit_type(self.circuit_type) \
            .with_metadata(circuit_name=circuit_name, system=system, signal=signal, wildcard={'MAGNET': wildcard}) \
            .signal_query() \
            .synchronize_time(timestamp_qds) \
            .convert_index_to_sec().dfs[0]

        return i_meas_diode_df, u_diode_rqd_df

    @execution_count()
    def query_current_voltage_diode_leads_pm(self, timestamp_fgc: int, results_table: pd.DataFrame) \
            -> List[Tuple[pd.DataFrame, pd.DataFrame, pd.DataFrame, pd.DataFrame]]:
        """ Method querying PM for current and voltage of a diode

        :param timestamp_fgc: FGC PM event timestamp (ns precision) to which the query is synchronised
        :type timestamp_fgc: int
        :param results_table: analysis results table
        :type results_table: pd.DataFrame
        :return: List of tuples of current and voltage pd.DataFrames
        """
        i_meas_u_diode_u_ref_pm_dfs = []
        for index, row in tqdm(results_table.iterrows(), total=results_table.shape[0], desc='Querying U, I Diode PM'):
            source_qds, timestamp_qds, u_diode_db = row[['magnet', 'timestamp_iqps', 'nqps_db']]

            if u_diode_db == 'PM':
                source_timestamp_nqps_df = self.find_source_timestamp_nqps_pm(source_qds, timestamp_qds)

                # ToDo: cover with a decorator
                if not source_timestamp_nqps_df.empty:
                    timestamp_nqps_board_a = source_timestamp_nqps_df.loc[0, 'timestamp']
                    timestamp_nqps_board_b = source_timestamp_nqps_df.loc[1, 'timestamp']
                    i_meas_u_diode_u_ref_pm_df = self._query_i_u_diode_pm(self.circuit_name, source_qds, timestamp_fgc,
                                                                          timestamp_nqps_board_a,
                                                                          timestamp_nqps_board_b)
                else:
                    i_meas_u_diode_u_ref_pm_df = (pd.DataFrame, pd.DataFrame, pd.DataFrame, pd.DataFrame)
            else:
                i_meas_u_diode_u_ref_pm_df = (pd.DataFrame, pd.DataFrame, pd.DataFrame, pd.DataFrame)

            i_meas_u_diode_u_ref_pm_dfs.append(i_meas_u_diode_u_ref_pm_df)

        return i_meas_u_diode_u_ref_pm_dfs

    @execution_count()
    def query_current_voltage_diode_leads_nxcals(self, results_table: pd.DataFrame, *, spark,
                                                 duration=[(50, 's'), (150, 's')]) -> List[Tuple[pd.DataFrame, pd.DataFrame]]:
        """ Method querying NXCALS for current and voltage of a diode

        :param results_table: analysis results table
        :type results_table: pd.DataFrame
        :param spark: spark session variable needed to perform an NXCALS query
        :type spark: spark session
        :return: List of tuples of current and voltage pd.DataFrames
        """
        i_meas_u_diode_nxcals_dfs = []
        for index, row in tqdm(results_table.iterrows(), total=results_table.shape[0],
                               desc='Querying U, I Diode leads NXCALS'):
            source_qds, timestamp_qds, u_diode_db = row[['magnet', 'timestamp_iqps', 'nqps_db']]

            timestamp_qds = Time.to_unix_timestamp(timestamp_qds)
            i_meas_u_diode_nxcals_df = self._query_i_u_diode_nxcals(self.circuit_name, source_qds,
                                                                    timestamp_qds, spark=spark, duration=duration)

            i_meas_u_diode_nxcals_dfs.append(i_meas_u_diode_nxcals_df)

        return i_meas_u_diode_nxcals_dfs

    # U_DIODE, U_EARTH
    @execution_count()
    def query_voltage_cals(self, system: str, signal: str, timestamp_sync: int, *, ldb,
                           duration=[(50, 's'), (150, 's')],
                           ) -> List[pd.DataFrame]:
        """ Method querying CALS for voltage signals (either U_DIODE or U_EARTH)
        Signals are synchronized to the input timestamp

        :param system: name of a leads system to query according to metadata
        :type system: str
        :param signal: name of leads signal to query according to metadata
        :param timestamp_sync: FGC PM event timestamp (ns precision) to which the query is synchronised
        :type timestamp_sync: int
        :param ldb: CALS pytimber connector
        :type ldb: pytimber
        :param duration: search duration before and after the start time
        :type duration: List[Tuple[int, str]]
        :return: list of pd.DataFrames (if a signal was not queried, an empty pd.DataFrame is returned and warning
            containing the raw query is returned)
        """
        return QueryBuilder().with_cals(ldb) \
            .with_duration(t_start=timestamp_sync, duration=duration) \
            .with_circuit_type(self.circuit_type) \
            .with_metadata(circuit_name=self.circuit_name, system=system, signal=signal, wildcard={'MAGNET': '*'}) \
            .signal_query(verbose=self.verbose) \
            .synchronize_time(timestamp_sync / 1e9).dfs

    # LEADS
    @execution_count()
    @check_nan_timestamp(return_type=pd.DataFrame())
    def find_timestamp_leads(self, t_start: int, system='LEADS', duration=[(10, 's'), (400, 's')],
                             ) -> pd.DataFrame:
        """ Method searching PM leads events
        If the timestamp is NaN, then an empty dataframe is returned.

        :param t_start: FGC PM event timestamp (ns precision) to which the query is synchronised
        :type t_start: int
        :param system: name of a leads system to query according to metadata
        :type system: str
        :param duration: search duration before and after the start time
        :type duration: List[Tuple[int, str]]
        :return: pd.DataFrame with source and timestamp of the PM LEADS events
        """
        return QueryBuilder().with_pm() \
            .with_duration(t_start=t_start, duration=duration) \
            .with_circuit_type(self.circuit_type) \
            .with_metadata(circuit_name=self.circuit_name, system=system) \
            .event_query(verbose=self.verbose).df

    @execution_count()
    @check_nan_timestamp_signals(return_type=pd.DataFrame())
    def query_leads(self, timestamp_fgc: int, source_timestamp_leads_df: pd.DataFrame, *, system='LEADS',
                    signal_names: List[str], spark, duration=[(30, 's'), (90, 's')]) -> List[pd.DataFrame]:
        """ Method querying leads signals from either PM or NXCALS
        If any timestamp is NaN, then a list of empty dataframes is returned.
        Signals are synchronized to input timestamp and index is converted to seconds.

        :param timestamp_sync: FGC PM event timestamp (ns precision) to which the query is synchronised
        :type timestamp_sync: int
        :param source_timestamp_leads_df:
        :param system: name of a leads system to query according to metadata
        :type system: str
        :param signal_names: list of signal names to query according to metadata description
        :type signal_names: List[str]
        :param spark: spark session variable needed to perform an NXCALS query
        :type spark: spark session
        :param duration: search duration before and after the start time
        :type duration: List[Tuple[int, str]]
        :return: list of pd.DataFrames (if a signal was not queried, an empty pd.DataFrame is returned and warning
            containing the raw query is returned)
        """
        if source_timestamp_leads_df.empty:
            return self._query_leads_nxcals(duration, signal_names, spark, system, timestamp_fgc)
        else:
            return self._query_leads_pm(signal_names, source_timestamp_leads_df, system, timestamp_fgc)

    def _query_leads_nxcals(self, duration: List[Tuple[int, str]], signal_names: List[str], spark, system: str,
                            timestamp_fgc: int) -> List[pd.DataFrame]:
        """ Method querying NXCALS for leads signals
        Signals are synchronized to input timestamp and index is converted to seconds.

        :param duration: search duration before and after the start time
        :type duration: List[Tuple[int, str]]
        :param signal_names: list of signal names to query according to metadata description
        :type signal_names: List[str]
        :param spark: spark session variable needed to perform an NXCALS query
        :type spark: spark session
        :param system: name of a leads system to query according to metadata
        :type system: str
        :param timestamp_fgc: FGC PM event timestamp (ns precision) to which the query is synchronised
        :type timestamp_fgc: int
        :return: list of pd.DataFrames (if a signal was not queried, an empty pd.DataFrame is returned and warning
            containing the raw query is returned)
        """
        return QueryBuilder().with_nxcals(spark) \
            .with_duration(t_start=timestamp_fgc, duration=duration) \
            .with_circuit_type(self.circuit_type) \
            .with_metadata(circuit_name=self.circuit_name, system=system, signal=signal_names) \
            .signal_query(verbose=self.verbose) \
            .synchronize_time(timestamp_fgc) \
            .convert_index_to_sec().dfs

    def _query_leads_pm(self, signal_names: List[str], source_timestamp_leads_df: pd.DataFrame, system: str,
                        timestamp_fgc: int) -> List[pd.DataFrame]:
        """ Method querying PM for leads signals.
        Signals are synchronized to input timestamp and index is converted to seconds.

        :param signal_names: list of signal names to query according to metadata description
        :type signal_names: List[str]
        :param source_timestamp_leads_df:
        :param system: name of a leads system to query according to metadata
        :type system: str
        :param timestamp_fgc: FGC PM event timestamp (ns precision) to which the query is synchronised
        :type timestamp_fgc: int
        :return: list of pd.DataFrames (if a signal was not queried, an empty pd.DataFrame is returned and warning
            containing the raw query is returned)
        """
        source_leads = source_timestamp_leads_df.loc[0, 'source']
        timestamp_leads = source_timestamp_leads_df.loc[0, 'timestamp']
        return QueryBuilder().with_pm() \
            .with_timestamp(timestamp_leads) \
            .with_circuit_type(self.circuit_type) \
            .with_metadata(circuit_name=self.circuit_name, system=system,
                           signal=signal_names, source=source_leads) \
            .signal_query(verbose=self.verbose) \
            .synchronize_time(timestamp_fgc) \
            .convert_index_to_sec().dfs

    # DFB
    def query_dfb_signal_nxcals(self, t_start, t_end, *, system, signal_names, spark):
        return QueryBuilder().with_nxcals(spark) \
            .with_duration(t_start=t_start, t_end=t_end) \
            .with_circuit_type(self.circuit_type) \
            .with_metadata(circuit_name=self.circuit_name, system=system, signal=signal_names) \
            .signal_query() \
            .synchronize_time() \
            .convert_index_to_sec() \
            .filter_median() \
            .dfs

    # Busbar
    @execution_count()
    def query_busbar_features_nxcals(self, t_start, t_end=None, duration=None, *, signal='U_RES',
                                     features=['mean', 'std', 'max', 'min', 'count'], spark):
        t_start, t_end = Time.get_query_period_in_unix_time(t_start, t_end, duration, '')
        return QueryBuilder() \
            .with_nxcals(spark) \
            .with_duration(t_start=t_start, t_end=t_end) \
            .with_circuit_type(self.circuit_type) \
            .with_metadata(circuit_name=self.circuit_name, system='BUSBAR', signal=signal, wildcard={'BUSBAR': '*'}) \
            .feature_query(features) \
            .calculate_max_abs() \
            .sort_busbar_location(self.circuit_type, self.circuit_name) \
            .df

    @execution_count()
    def calculate_current_plateau_start_end(self, t_start, t_end, *, i_meas_threshold=50, min_duration_in_sec=1, spark):
        i_meas_raw_dfs = QueryBuilder().with_nxcals(spark) \
            .with_duration(t_start=t_start, t_end=int(t_end)) \
            .with_circuit_type(self.circuit_type) \
            .with_metadata(circuit_name=self.circuit_name, system='PC', signal='I_MEAS') \
            .signal_query() \
            .dfs

        # Find plateau start and end based on the current profile
        plateau_starts = []
        plateau_ends = []
        for i_meas_raw_df in i_meas_raw_dfs:
            if i_meas_raw_df.max().values[0] < 100:
                print('There is at least one signal with the peak current below 100 A.')
                return pd.DataFrame()

        for i_meas_raw_df in i_meas_raw_dfs:
            plateau_start, plateau_end = find_plateau_start_and_end(i_meas_raw_df, i_meas_threshold=i_meas_threshold,
                                                                    min_duration_in_sec=min_duration_in_sec)
            if plateau_start is None:
                print('\tThe signal does not contain ramp-up!')
                return pd.DataFrame

            plateau_starts.append(plateau_start)
            plateau_ends.append(plateau_end)

        max_len_plateau_starts = max(len(plateau_start) for plateau_start in plateau_starts)
        max_len_plateau_ends = max(len(plateau_end) for plateau_end in plateau_ends)
        if max_len_plateau_starts != max_len_plateau_ends:
            return pd.DataFrame()

        # Take only the latest plateau start and the earliest plateau ends
        return extract_intersecting_plateaus(plateau_ends, plateau_starts)

    @execution_count()
    def get_busbar_resistances(self, t_start, t_end, plateau_start, plateau_end, *, signal_name='U_RES',
                               system='BUSBAR', spark):

        plateau_start = np.array(plateau_start)
        plateau_end = np.array(plateau_end)
        ramp_up_start, ramp_up_end = find_start_end_of_the_longest_ramp_up(plateau_start, plateau_end)

        def translate(timestamp):
            mask_plateau = (timestamp >= plateau_start) & (timestamp <= plateau_end)
            index_plateau = np.where(mask_plateau == True)[0]

            if len(index_plateau) > 0:
                return int((index_plateau[0] + 1))

            if (timestamp > ramp_up_start) and (timestamp < ramp_up_end):
                return 0

            return -1

        plateau_translate_udf = F.udf(translate, IntegerType())

        # query current
        i_meas_feature_df = QueryBuilder() \
            .with_nxcals(spark) \
            .with_duration(t_start=int(t_start), t_end=int(t_end)) \
            .with_circuit_type(self.circuit_type) \
            .with_metadata(circuit_name=self.circuit_name, system='PC', signal='I_MEAS') \
            .feature_query(['mean', 'std'], function=plateau_translate_udf) \
            .sort_values(by='class') \
            .df

        # query voltage
        u_res_feature_df = QueryBuilder() \
            .with_nxcals(spark) \
            .with_duration(t_start=int(t_start), t_end=int(t_end)) \
            .with_circuit_type(self.circuit_type) \
            .with_metadata(circuit_name=self.circuit_name, system=system,
                           signal=signal_name, wildcard={'BUSBAR': '*'}) \
            .feature_query(['mean', 'std'], function=plateau_translate_udf) \
            .sort_busbar_location(self.circuit_type, circuit_name=self.circuit_name) \
            .correct_voltage_sign() \
            .df

        return u_res_feature_df, i_meas_feature_df

    @staticmethod
    def query_and_plot_outlier_voltage(res_outliers_df, t_start, t_end, t_0, plateau_start, plateau_end, *,
                                       is_automatic=False, spark):
        if not res_outliers_df.empty:
            print('Resistance(s) outside of range:')
            display(HTML(res_outliers_df.to_html()))

        for index, row in res_outliers_df.iterrows():

            u_df = QueryBuilder().with_nxcals(spark) \
                .with_duration(t_start=t_start, t_end=t_end) \
                .with_query_parameters(nxcals_system='CMW', signal=row['nxcals_variable_name']) \
                .signal_query() \
                .synchronize_time() \
                .convert_index_to_sec().dfs[0]

            if not u_df.empty:
                ax = u_df.plot(figsize=(15, 7), grid=True)
                # Add ranges of detected plateaus in orange
                for ps, pe in zip(plateau_start, plateau_end):
                    ax.axvspan(xmin=(ps - Time.to_unix_timestamp(t_0)) / 1e9,
                               xmax=(pe - Time.to_unix_timestamp(t_0)) / 1e9,
                               facecolor='xkcd:goldenrod')
                ax.set_xlabel("time, [s]", fontsize=15)
                ax.set_ylabel("U, [V]", fontsize=15)
                ax.tick_params(labelsize=15)
                plt.show()

            check_show_next(index, len(res_outliers_df), is_automatic=is_automatic)
