import warnings
from typing import Dict

import numpy as np
import pandas as pd

from lhcsmapi.analysis.CircuitAnalysis import CircuitAnalysis
from lhcsmapi.reference.Reference import Reference


class IpdCircuitAnalysis(CircuitAnalysis):

    # Timestamp table
    def create_timestamp_table(self, timestamp_dct: Dict[str, int], circuit_name='') -> None:
        """ Method creating timestamp table as well as checking the timing between considered timestamps.

        :param timestamp_dct: a dictionary from timestamp name to its value (it is assumed that does not contain NaN)
        :type timestamp_dct: Dict[str, int]
        :param circuit_name: name of a circuit
        :type circuit_name: str
        """
        super().create_timestamp_table(timestamp_dct)

        if ('QDS_A' not in timestamp_dct) or np.isnan(timestamp_dct['QDS_A']):
            warnings.warn('QDS Board A not present in PM buffer!', stacklevel=2)
        elif ('QDS_B' not in timestamp_dct) or np.isnan(timestamp_dct['QDS_B']):
            warnings.warn('QDS Board B not present in PM buffer!', stacklevel=2)
        else:
            if timestamp_dct['QDS_B'] - timestamp_dct['QDS_A'] != int(1e6):
                warnings.warn('QDS PM events (board A and B) are not synchronized to 1 ms!')

        if ('QDS_A' in timestamp_dct) and ('FGC' in timestamp_dct):
            if abs(timestamp_dct['QDS_A'] - timestamp_dct['FGC']) > int(20e6):
                warnings.warn('QDS_A and FGC PM events are not synchronized to +/- 20 ms!')
        else:
            warnings.warn('Either QDS_A or FGC PM event is missing. '
                          'Can not check synchronization of QDS and FGC timestamps!')

    def create_mp3_results_table(self):
        """ Method creating an MP3-compatible quench database row for an IPD circuit

        :return: pd.DataFrame with the results of analysis calculation
        """
        columns_mp3 = Reference.read_quench_database_columns(circuit_type=self.circuit_type, table_type='MP3')
        mp3_table = pd.DataFrame(columns=columns_mp3, index=self.results_table.index)

        # Union of MP3 and results table columns
        columns_lhcsm = Reference.read_quench_database_columns(circuit_type=self.circuit_type, table_type='LHCSM')
        same_columns = list(set(columns_mp3).intersection(set(columns_lhcsm)))

        mp3_table[same_columns] = self.results_table[same_columns]

        # Fill non-matching column names
        mp3_table['Quench Current'] = self.results_table['I_MEAS']
        mp3_table['DCCT MIITs'] = self.results_table['MIITs_DCCT']
        if not np.isnan(self.results_table['dUres/dt'].values[0]):
            mp3_table['dt_thres'] = 0.1 / self.results_table['dUres/dt'].values[0]
        mp3_table['Person who analyzed'] = self.results_table['performed_by']

        return mp3_table
