from typing import Union, List, Tuple

import pandas as pd

import lhcsmapi
from lhcsmapi.Time import Time
from lhcsmapi.analysis.CircuitQuery import CircuitQuery, execution_count
from lhcsmapi.metadata.SignalMetadata import SignalMetadata
from lhcsmapi.analysis.decorators import check_nan_timestamp
from lhcsmapi.pyedsl.QueryBuilder import QueryBuilder
from lhcsmapi.reference.Reference import Reference


class IpdCircuitQuery(CircuitQuery):

    def __init__(self, circuit_type, circuit_name, max_executions=None, verbose=True):
        super().__init__(circuit_type, circuit_name, max_executions, verbose)

        if self.circuit_name in SignalMetadata.get_circuit_names('IPD2'):
            self.circuit_type = 'IPD2'
        elif self.circuit_name in SignalMetadata.get_circuit_names('IPD2_B1B2'):
            self.circuit_type = 'IPD2_B1B2'
        else:
            raise ValueError('Circuit name %s does not map to internal metadata.' % self.circuit_name)

    def create_report_analysis_template(self, timestamp_fgc: int, author='') -> pd.DataFrame:
        """ Method creating a report analysis template for IPD circuits

        :param timestamp_fgc: timestamp of an FGC event
        :type timestamp_fgc: int
        :param author: NICE account name of the analysis author
        :type author: str
        :return: pre-filled pd.DataFrame with a template to be completed further with data from an analysis of
            a powering event in IT circuits
        """
        columns = Reference.read_quench_database_columns(circuit_type=self.circuit_type, table_type='LHCSM')
        results_table = pd.DataFrame(columns=columns, index=[0])

        results_table['Circuit Name'] = self.circuit_name
        results_table['Date'] = Time.to_string_short(timestamp_fgc).split(' ')[0]
        results_table['Time'] = Time.to_string_short(timestamp_fgc).split(' ')[1]
        results_table['performed_by'] = author

        results_table['lhcsmapi_version'] = lhcsmapi.__version__

        return results_table

    @execution_count()
    @check_nan_timestamp(return_type=pd.DataFrame())
    def find_source_timestamp_qds(self, t_start: Union[int, str, float], duration=[(10, 's'), (400, 's')],
                                  ) -> pd.DataFrame:
        """ Method searching PM QDS events
        If the start is NaN, then an empty dataframe is returned.

        :param t_start: start time of a search
        :type t_start: Union[str, int, float]
        :param duration: search duration before and after the start time
        :type duration: List[Tuple[int, str]]
        :return: pd.DataFrame with source and timestamp of the PM QDS events
        """
        return QueryBuilder().with_pm() \
            .with_duration(t_start=t_start, duration=duration) \
            .with_circuit_type(self.circuit_type) \
            .with_metadata(circuit_name=self.circuit_name, system='QDS', source='*') \
            .event_query(verbose=self.verbose) \
            .sort_values(by='timestamp').df
