from typing import Tuple, List

import numpy as np
import pandas as pd

import lhcsmapi
from lhcsmapi.Time import Time
from lhcsmapi.metadata.SignalMetadata import SignalMetadata
from lhcsmapi.analysis.CircuitQuery import CircuitQuery, execution_count
from lhcsmapi.pyedsl.QueryBuilder import QueryBuilder
from lhcsmapi.reference.Reference import Reference


class ItCircuitQuery(CircuitQuery):
    """Class extending the CircuitQuery class with methods for creation of templates of analysis tables in IT circuits.

    """

    def create_report_analysis_template(self, timestamp_fgc: int, author='') -> pd.DataFrame:
        """ Method creating a report analysis template for IT circuits

        :param timestamp_fgc: timestamp of an FGC event
        :type timestamp_fgc: int
        :param author: NICE account name of the analysis author
        :type author: str
        :return: pre-filled pd.DataFrame with a template to be completed further with data from an analysis of
            a powering event in IT circuits
        """
        columns = Reference.read_quench_database_columns(circuit_type='IT', table_type='LHCSM')
        results_table = pd.DataFrame(columns=columns, index=[0])

        results_table['Circuit Name'] = self.circuit_name
        results_table['Date'] = Time.to_string_short(timestamp_fgc).split(' ')[0]
        results_table['Time'] = Time.to_string_short(timestamp_fgc).split(' ')[1]
        results_table['performed_by'] = author

        results_table['lhcsmapi_version'] = lhcsmapi.__version__

        return results_table

    @staticmethod
    def split_source_timestamp_fgc(source_timestamp_fgc_df: pd.DataFrame, pc_name: str) -> Tuple[str, int]:
        """ Method splitting source and timestamp for a given source and timestamp dataframe as well as IT power
        converter name.

        :param source_timestamp_fgc_df: a dataframe with FGC PM source and timestamp obtained with QueryBuilder
        :type source_timestamp_fgc_df: pd.DataFrame
        :param pc_name: name of an IT circuit power converter for which the source and timestamp is retrieved
        :type pc_name: str
        :return: a tuple with FGC PM event source and timestamp
        """
        source_timestamp_fgc_with_pc_name_df = source_timestamp_fgc_df[
            source_timestamp_fgc_df['source'].str.contains(pc_name)]
        if source_timestamp_fgc_with_pc_name_df.empty:
            return '', np.nan
        else:
            source = source_timestamp_fgc_with_pc_name_df['source'].values[0]
            timestamp = source_timestamp_fgc_with_pc_name_df['timestamp'].values[0]
            return source, timestamp

    @execution_count()
    def query_pc_pm_with_source(self, timestamp_fgc: int, timestamp_sync: int, source: str, *,
                                signal_names: List[str]) -> List[pd.DataFrame]:
        """ Method querying PM for PC signals with a given source and a list of signal names.

        :param timestamp_fgc: timestamp of an FGC event (ns precision)
        :type timestamp_fgc: int
        :param timestamp_sync: timestamp to which the signals are synchronised, typically FGC timestamp,
            (ns precision)
        :type timestamp_sync: int
        :param source: PM FGC source for which the query is executed
        :type source: str
        :param signal_names: list of signal names to query according to metadata description
        :type signal_names: List[str]
        :return: list of pd.DataFrame with synchronized signals there are as many elements as the signal names
        """

        # Get metadata
        md_pc_pm = SignalMetadata.get_circuit_signal_database_metadata(self.circuit_type, self.circuit_name, 'PC', 'PM')

        # Get signal names to query - using general-purpose query
        if any([signal_name not in md_pc_pm for signal_name in signal_names]):
            raise ValueError('At least one signal name is not in the metadata, please correct your input!')

        signal_names_pm = [md_pc_pm[signal_name] for signal_name in signal_names]

        # Query and append suffix to columns
        suffix = '_' + source.split('.')[-2]
        return QueryBuilder().with_pm() \
            .with_timestamp(timestamp_fgc) \
            .with_query_parameters(system=md_pc_pm['system'], source=source, className=md_pc_pm['className'],
                                   signal=signal_names_pm) \
            .signal_query(verbose=self.verbose) \
            .synchronize_time(timestamp_sync) \
            .convert_index_to_sec() \
            .append_suffix_to_columns(suffix) \
            .dfs
