import math
from copy import deepcopy

import pandas as pd
import matplotlib.pyplot as plt
from IPython.display import display, HTML

from lhcsmapi.Time import Time
from lhcsmapi.analysis.CircuitAnalysis import CircuitAnalysis
from lhcsmapi.analysis.expert_input import check_show_next, get_expert_decision, get_expert_comment
import lhcsmapi.analysis.qds.QdsRbAnalysis as QdsRbAnalysis
from lhcsmapi.analysis.qh.QuenchHeaterRbAnalysis import analyze_single_qh
from lhcsmapi.metadata.SignalMetadata import SignalMetadata
from lhcsmapi.pyedsl.AssertionBuilder import AssertionBuilder
from lhcsmapi.reference.Reference import Reference
from lhcsmapi.metadata.MappingMetadata import MappingMetadata


class RbCircuitAnalysis(CircuitAnalysis):

    # EE
    def analyze_ee_temp_and_cooling(self, t_res_body_long_dfs, st_res_overtemp_long_dfs, *, t_greater, value_max):
        is_signal_correct = AssertionBuilder().with_signal(t_res_body_long_dfs) \
            .for_time_greater_than(t_greater=t_greater) \
            .has_min_max_value(value_max=value_max) \
            .show_plot(y_label='T_RES_BODY, [degC]',
                       data_sec=st_res_overtemp_long_dfs, drawstyle_sec='steps-post',
                       ylabel_sec='ST_RES_OVERTEMP, [-]') \
            .get_assertion_result()

        if self.results_table is not None:
            analysis = '%s_EE_T_RES_BODY_2_HOURS_AFTER_FPA' % self.circuit_type
            status = 'Within reference' if is_signal_correct else 'Outside of reference'
            self.results_table[analysis] = get_expert_comment(analysis, self.is_automatic, status)

    # QH
    def analyze_qh(self, u_hds_dfs, i_hds_dfs, u_hds_ref_dfs, i_hds_ref_dfs):
        index_max = self.results_table.index[-1]
        for index, row in self.results_table.iterrows():
            source_qh = row['magnet']
            timestamp_qh = row['timestamp_iqps'] + 1000000
            print('{}/{}: Analysing quench heater signals of {} on {}, {}' \
                  .format(index, index_max, source_qh, Time.to_string(timestamp_qh), timestamp_qh))

            is_qh_correct = analyze_single_qh(timestamp_qh, source_qh, u_hds_dfs[index], i_hds_dfs[index],
                                              u_hds_ref_dfs[index],i_hds_ref_dfs[index])

            # Update results table
            if self.results_table is not None:
                status = 'Within reference' if is_qh_correct else 'Outside of reference'
                self.results_table.loc[index, 'QH'] = status

            # Show next
            check_show_next(index, index_max, self.is_automatic)
    # QDS
    def analyze_qds(self, circuit_name, timestamp_fgc, timestamp_pic, u_qds_dfs, u_nqps_dfs, i_meas_df):
        index_max = self.results_table.index[-1]
        for index, row in self.results_table.iterrows():
            source_qds, source_nqps, timestamp_qds, timestamp_nqps, i_quench, nqps_db = \
                row[['magnet', 'nqps_crate', 'timestamp_iqps', 'timestamp_nqps', 'I_MEAS_quench', 'nqps_db']]

            print('{}/{}: Analysing quench detection signals of {} on {}, {}' \
                  .format(index, self.results_table.index[-1], source_qds, Time.to_string(timestamp_qds), timestamp_qds))

            u_qs0_df, u_1_df, u_2_df, st_nqd0_df, st_magnet_ok_df = u_qds_dfs[index]

            # Analysis
            report_qds_trigger = QdsRbAnalysis.analyze_qds_trigger(u_qs0_df, st_magnet_ok_df, st_nqd0_df)
            t_start_quench = report_qds_trigger.at[0, 't_start_quench [s]']
            t_st_nqd0 = report_qds_trigger.at[0, 't_st_nqd0 [s]']
            t_st_magnet_ok = report_qds_trigger.at[0, 't_st_magnet_ok [s]']

            # # iQPS
            nqps_crate = MappingMetadata.get_crate_name_from_magnet_name(circuit_type='RB', magnet=source_qds)
            dt_from_first_quench = (timestamp_qds - self.results_table.at[0, 'timestamp_iqps'])/1e9
            title = "Magnet: {}, iQPS Time Stamp: {}, nQPS Crate: {}, Current = {} A, Time from the 1st quench {} s". \
                format(source_qds, Time.to_string_short(timestamp_qds), nqps_crate, i_quench, dt_from_first_quench)
            fig, ax = plt.subplots(2, 2, figsize=(30, 15))
            fig.suptitle(title, fontsize=20)
            xlim = (t_start_quench - 0.025, t_st_magnet_ok + 0.075)
            QdsRbAnalysis.plot_u_qds(u_qs0_df, u_1_df, u_2_df, timestamp_qds, timestamp_pic, title=title,
                                     xmin=t_start_quench, xmax=t_st_nqd0, xlim=xlim, ax=ax[0, 0])
            QdsRbAnalysis.plot_st_qds(u_qs0_df, st_magnet_ok_df, st_nqd0_df, xmin=t_start_quench, xmax=t_st_nqd0,
                                      title=title, xlim=xlim, ax=ax[1, 0])

            if nqps_db == 'PM':
                u_nqps_shift_df = QdsRbAnalysis.synchronize_nqps_to_iqps(u_nqps_dfs[index], timestamp_nqps,
                                                                         timestamp_qds)
                u_nqps_shift_df = [u_nqps_shift_df_el for u_nqps_shift_df_el in u_nqps_shift_df if
                                   not u_nqps_shift_df_el.empty]
                u_nqps_shift_diff_df = QdsRbAnalysis.calculate_col_diff(u_nqps_shift_df)
                quenched_magnets = self.results_table.loc[:index - 1, 'magnet'].values
                u_nqps_shift_diff_filt_dfs = QdsRbAnalysis.find_dfs_with_quenched_magnet(u_nqps_shift_diff_df,
                                                                                         source_qds)
                t_nqps_diff_max, u_nqps_diff_max = QdsRbAnalysis.get_max_diff_val_and_idx(u_nqps_shift_diff_filt_dfs,
                                                                                          t_st_magnet_ok,
                                                                                          quenched_magnets)

                QdsRbAnalysis.plot_u_diode_pm_nqds(u_nqps_shift_df, timestamp_nqps, circuit_name, ax=ax[0, 1])
                # Plot only the differences with the quenched magnet
                if u_nqps_shift_diff_filt_dfs:
                    QdsRbAnalysis.plot_u_diode_diff_pm(u_nqps_shift_diff_filt_dfs, timestamp_nqps, circuit_name,
                                                       t_st_magnet_ok, ax=ax[1, 1])
                else:
                    ax[1, 1].axis('off')

                if (u_nqps_diff_max > 1) & (t_nqps_diff_max < t_st_magnet_ok):
                    is_iqps_or_nqps = 'nQPS'
                else:
                    is_iqps_or_nqps = 'iQPS'
            else:
                u_nqps_shift_df = u_nqps_dfs[index]

                i_meas_shift_df = deepcopy(i_meas_df)
                i_meas_shift_df.index = i_meas_shift_df.index + (timestamp_fgc - timestamp_qds) * 1e-9
                QdsRbAnalysis.plot_u_diode_cals(u_nqps_shift_df, timestamp_nqps, circuit_name,
                                                i_meas_df=i_meas_shift_df, ax=ax[0, 1])
                QdsRbAnalysis.plot_u_diode_cals(u_nqps_shift_df, timestamp_nqps, circuit_name, i_meas_df=None,
                                                ax=ax[1, 1], xlim=[-2, 2])

                if 'nQPS' in self.results_table.at[index, 'comments']:
                    is_iqps_or_nqps = 'nQPS'
                else:
                    is_iqps_or_nqps = '?'

                t_nqps_diff_max = math.nan
                u_nqps_diff_max = math.nan

            ax[0, 1].xaxis.set_tick_params(which='both', labelbottom=True)
            plt.show()

            report_qds_trigger['t_nqps_diff_max [s]'] = t_nqps_diff_max
            report_qds_trigger['u_nqps_diff_max [V]'] = u_nqps_diff_max
            display(HTML(report_qds_trigger.to_html()))

            du_dt = report_qds_trigger.at[0, 'du_dt [V/s]']
            self.results_table.at[index, 'dU_QS0/dt'] = du_dt
            self.results_table.at[index, 'U_QS0_polarity'] = 'Pos' if du_dt > 0 else 'Neg'
            self.results_table.at[index, 'comments'] = report_qds_trigger.at[0, 'analysis_comment']
            if self.is_automatic:
                self.results_table.at[index, 'iqps_or_nqps_trigger'] = is_iqps_or_nqps
            else:
                self.results_table.at[index, 'iqps_or_nqps_trigger'] = \
                    get_expert_decision('What is the source of trigger', ['iQPS', 'nQPS'])
                self.results_table.at[index, 'quench_type'] = \
                    get_expert_decision('What is the quench type', ['EM', 'BEAM', 'Coupling', 'Training', 'HP'])

            # Show next
            check_show_next(index, index_max, self.is_automatic)

    def create_mp3_results_table(self, circuit_name, timestamp_fgc):

        columns_mp3 = Reference.read_quench_database_columns(circuit_type='RB', table_type='MP3')
        mp3_table_df = pd.DataFrame(columns=columns_mp3, index=self.results_table.index)

        def create_pm_file_name(timestamp, name):
            return Time.to_string_short(timestamp).replace(' ', '_').replace(':', '').replace('-', '') + '_' + name

        # Circuit
        mp3_table_df['Circuit'] = self.results_table['circuit']
        # Date
        mp3_table_df['Date'] = Time.to_string_short(timestamp_fgc).split(' ')[0]
        # Time
        mp3_table_df['Time'] = Time.to_string_short(timestamp_fgc).split(' ')[1].split('.')[0]
        # FGC PM file
        pc_name = SignalMetadata.get_circuit_name_metadata(self.circuit_type, circuit_name)['PC']['%PC%']
        mp3_table_df['FGC PM file'] = create_pm_file_name(timestamp_fgc, pc_name)
        # Position
        mp3_table_df['Position'] = self.results_table['magnet']
        # QPS PM file
        mp3_table_df['QPS PM file'] = self.results_table['timestamp_iqps']
        mp3_table_df['QPS PM file'] = mp3_table_df.apply(
            lambda row: create_pm_file_name(row['QPS PM file'], row['Position']), axis=1)
        # flagged in APEX
        mp3_table_df['flagged in APEX'] = 'no'
        # iQPS or nQPS
        mp3_table_df['iQPS or nQPS'] = self.results_table['iqps_or_nqps_trigger']
        # I_q
        mp3_table_df['I_q'] = self.results_table['I_MEAS_quench']
        # Nr in Q event
        mp3_table_df['Nr in Q event'] = mp3_table_df.apply(lambda row: row.name + 1, axis=1)
        # Magnet ID
        mp3_table_df['Magnet ID'] = self.results_table['magnet_id']
        # QS0 polarity
        mp3_table_df['QS0 polarity'] = self.results_table['U_QS0_polarity']
        # dV(QS0)/dt_aver
        mp3_table_df['dV(QS0)/dt_aver'] = self.results_table['dU_QS0/dt']
        # R_leads_max
        mp3_table_df['R_leads_max'] = self.results_table['R_DIODE_LEADS_RB']
        # Comments
        mp3_table_df['Comments'] = self.results_table['comments']
        # I_MEAS
        mp3_table_df['I_MEAS'] = self.results_table['RB_PC_I_MEAS_tau']
        # nQPS
        mp3_table_df['nQPS'] = self.results_table['DIODE_RB_U_DIODE_RB']
        # Heater
        mp3_table_df['Heater'] = self.results_table['QH']
        # EE
        mp3_table_df['EE'] = self.results_table['RB_EE_U_DUMP_RES_tau_res']
        # V feeler
        mp3_table_df['V feeler'] = self.results_table['VF_U_EARTH_RB']
        # Analysis performed by
        mp3_table_df['Analysis performed by'] = self.results_table['performed_by']

        return mp3_table_df
