from typing import List, Tuple

import requests
import numpy as np
import pandas as pd
from tqdm.notebook import tqdm

import lhcsmapi
from lhcsmapi.Time import Time
from lhcsmapi.analysis.CircuitQuery import CircuitQuery, get_quench_current, execution_count
from lhcsmapi.analysis.decorators import check_nan_timestamp_signals
from lhcsmapi.metadata.MappingMetadata import MappingMetadata
from lhcsmapi.metadata.SignalMetadata import SignalMetadata
from lhcsmapi.pyedsl.QueryBuilder import QueryBuilder
from lhcsmapi.reference.Reference import Reference


class RbCircuitQuery(CircuitQuery):
    """Class extending the CircuitQuery class with methods for creation of templates of analysis tables in RB circuits.

    """

    @execution_count()
    def create_report_analysis_template(self, source_timestamp_qds_df: pd.DataFrame,
                                        source_timestamp_nqps_df: pd.DataFrame,qps_board_type_df: pd.DataFrame,
                                        timestamp_fgc: int, i_meas_df: pd.DataFrame, author: str) -> pd.DataFrame:
        """ Method creating a report analysis template for main dipole circuits

        :param source_timestamp_qds_df: table with QDS PM source and timestamp
        :type source_timestamp_qds_df: pd.DataFrame
        :param source_timestamp_nqps_df: table with QDS PM source and timestamp
        :type source_timestamp_nqps_df: pd.DataFrame
        :param qps_board_type_df: table with QDS PM board type
        :type qps_board_type_df: pd.DataFrame
        :param timestamp_fgc: unix timestamp of an FGC event in RB circuit in nanoseconds
        :type timestamp_fgc: int
        :param i_meas_rqd_df: measured power converter current in RB circuit
        :type i_meas_rqd_df: pd.DataFrame
        :param author: NICE account name of the analysis author
        :type author: str
        :return: pre-filled pd.DataFrame with a template to be filled further with data from analysis
            of a powering event in main dipole (RB) circuits
        """

        columns = Reference.read_quench_database_columns(circuit_type='RB', table_type='LHCSM')
        report_table = pd.DataFrame(columns=columns)

        report_table['magnet'] = source_timestamp_qds_df['source']
        report_table['circuit'] = self.circuit_name
        report_table['magnet_id'] = report_table['magnet'].apply(
            lambda col: MappingMetadata.get_magnet_id_from_magnet_name(self.circuit_type, 'MB.{}'.format(col)))
        report_table['timestamp_iqps'] = source_timestamp_qds_df['timestamp']
        timestamp_qds_0 = report_table.loc[0, 'timestamp_iqps']
        report_table['dt_quench'] = report_table['timestamp_iqps'].apply(lambda col: (col - timestamp_qds_0) * 1e-9)
        report_table['datetime_iqps'] = report_table['timestamp_iqps'].apply(lambda col: Time.to_string_short(col))
        report_table['t_quench'] = report_table['timestamp_iqps'].apply(lambda col: (col - timestamp_fgc) * 1e-9)
        report_table['I_MEAS_quench'] = report_table['t_quench'].apply(lambda col: get_quench_current(i_meas_df, col))
        report_table['iqps_board_type'] = qps_board_type_df['iqps_board_type']
        report_table['nqps_crate'] = report_table['magnet'].apply(
            lambda col: MappingMetadata.get_crate_name_from_magnet_name(self.circuit_type, col))
        report_table['nqps_db'] = report_table['t_quench'].apply(lambda col: 'PM' if col < 2 else 'NXCALS')
        report_table['lhcsmapi_version'] = lhcsmapi.__version__
        report_table['performed_by'] = author
        report_table['comments'] = ''
        report_table.drop(columns='t_quench', inplace=True)

        source_timestamp_nqps_renamed_df = source_timestamp_nqps_df.rename(
            columns={'source': 'nqps_crate', 'timestamp': 'timestamp_nqps'})
        # ToDo If an nQPS PM event is missing, then NaN is set enforcing the entire row to be set to into float
        report_table = report_table.merge(source_timestamp_nqps_renamed_df[['timestamp_nqps', 'nqps_crate']],
                                      on='nqps_crate', how='left').sort_values(by='timestamp_iqps') \
            .reset_index(drop=True)
        report_table['datetime_nqps'] = report_table['timestamp_nqps'].apply(lambda col: Time.to_string_short(col))

        return report_table

    # PIC
    @execution_count()
    @check_nan_timestamp_signals(return_type=pd.DataFrame())
    def find_timestamp_pic(self, timestamp_fgc: int, *, spark: object,
                           duration: List[Tuple[int, str]] = [(1, 's'), (60, 's')]) -> List[int]:
        """ Method finding PIC timestamps by querying NXCALS database around an FGC timestamp.
        If the input timestamp is NaN, then a NaN is returned.
        If the PIC timestamp was not found, then a list of NaN is returned.

        :param timestamp_fgc: FGC PM event timestamp (ns precision)
        :type timestamp_fgc: int
        :param spark: spark session variable needed to perform an NXCALS query
        :type spark: spark session
        :param duration: search duration before and after the FGC PM timestamp
        :type duration: List[Tuple[int, str]]
        :return: list of float value of PIC timestamp in ns precision
        """
        st_abort_pic_dfs = QueryBuilder().with_nxcals(spark) \
            .with_duration(t_start=int(timestamp_fgc), duration=duration) \
            .with_circuit_type(self.circuit_type) \
            .with_metadata(circuit_name=self.circuit_name, system='PIC', signal='ST_ABORT_PIC') \
            .signal_query(verbose=self.verbose).dfs

        if all([st_abort_pic_df.empty for st_abort_pic_df in st_abort_pic_dfs]):
            return [np.nan, np.nan]

        return [st_abort_pic_df.index[0] for st_abort_pic_df in st_abort_pic_dfs]

    # iQPS
    @execution_count()
    def query_pm_iqps_board_type(self, source_timestamp_qds_df: pd.DataFrame) -> pd.DataFrame:
        """ Method querying PM for qps board type

        :param source_timestamp_qds_df: table with QDS PM source and timestamp
        :type source_timestamp_qds_df: pd.DataFrame

        :return: table with iQPS board types (1, 0)
        """
        iqps_board_type_df = pd.DataFrame(columns=['iqps_board_type'])
        metadata_qds = SignalMetadata.get_circuit_signal_database_metadata(self.circuit_type, self.circuit_name, 'QDS',
                                                                           'PM')
        board_types = []
        for index, row in tqdm(source_timestamp_qds_df.iterrows(), total=source_timestamp_qds_df.shape[0],
                               desc='Querying PM'):
            source_qds = row['source']
            timestamp_qds = row['timestamp']

            # ToDo: It would be nice to add this query to the library
            pm_request = "http://pm-api-pro/v2/pmdata?system=QPS&className=" + metadata_qds['className'] + "&source=" \
                         + source_qds + "&TimestampInNanos=" + str(timestamp_qds) + "&pageSize=1&pageIndex=0"
            resp = requests.get(pm_request)
            board_type = resp.json()["content"][0]["namesAndValues"][1]["value"]

            board_types.append(board_type)

        iqps_board_type_df['iqps_board_type'] = board_types
        return iqps_board_type_df

    @execution_count()
    def query_voltage_logic_iqps(self, results_table, *, signal_names):
        """ Method querying voltage and logic signals of iQPS.
        The signals are synchronized in time and converted from ns to s.

        :param results_table: table with source and timestamp for iQPS
        :type results_table: pd.DataFrame
        :param signal_names: list of signal names to query
        :type signal_names: List[str]
        :return: list of pd.DataFrames (if a signal was not queried, an empty pd.DataFrame is returned and warning
            containing the raw query is returned)
        """
        # ToDo: Should use source_timestamp_df instead of the results_table!
        u_qds_dfs = []
        for index, row in tqdm(results_table.iterrows(), total=results_table.shape[0],
                               desc='Querying iQPS logic PM'):
            source_qds, timestamp_qds = row[['magnet', 'timestamp_iqps']]

            u_qds_df = QueryBuilder().with_pm() \
                .with_timestamp(timestamp_qds) \
                .with_circuit_type(self.circuit_type) \
                .with_metadata(circuit_name=self.circuit_name, system='QDS', source=source_qds,
                               wildcard={'CELL': source_qds},
                               signal=signal_names) \
                .signal_query() \
                .synchronize_time(timestamp_qds) \
                .convert_index_to_sec().dfs

            u_qds_dfs.append(u_qds_df)

        return u_qds_dfs

    @execution_count()
    def query_voltage_nqps(self, results_table, *, ldb):
        """ Method querying voltage of nQPS from either PM or CALS.
            The signals are synchronized in time and converted from ns to s.

        :param results_table: table with source and timestamp for nQPS
        :type results_table: pd.DataFrame
        :param ldb: pytimber connector for CALS
        :type ldb: pytimber connector
        :return: list of pd.DataFrames (if a signal was not queried, an empty pd.DataFrame is returned and warning
            containing the raw query is returned)
        """
        u_nqps_dfs = []
        for index, row in tqdm(results_table.iterrows(), total=results_table.shape[0],
                               desc='Querying CALS U_DIODE_RB, U_REF_N1'):
            source_nqps, timestamp_nqps, timestamp_qds, nqps_db = row[['nqps_crate', 'timestamp_nqps',
                                                                        'timestamp_iqps', 'nqps_db']]

            magnet_names = MappingMetadata.get_magnet_names_from_crate_name(self.circuit_type, source_nqps)
            if nqps_db == 'PM':
                u_diode_dfs, u_ref_n1_dfs = self._query_nqps_pm(magnet_names, source_nqps, timestamp_nqps)
            else:
                u_diode_dfs, u_ref_n1_dfs = self._query_nqps_cals(source_nqps, magnet_names, timestamp_qds, ldb=ldb)

            u_nqps_dfs.append(u_diode_dfs + u_ref_n1_dfs)

        return u_nqps_dfs

    def _query_nqps_cals(self, source_nqps, magnet_names, timestamp_qds, *, ldb, duration=[(10, 's'), (350, 's')]):

        u_diode_dfs = QueryBuilder().with_cals(ldb) \
            .with_duration(t_start=timestamp_qds, duration=duration) \
            .with_circuit_type(self.circuit_type) \
            .with_metadata(circuit_name=self.circuit_name, system='DIODE_RB', source=source_nqps,
                           wildcard={'MAGNET': magnet_names}, signal='U_DIODE_RB') \
            .signal_query() \
            .synchronize_time(timestamp_qds / 1e9) \
            .dfs
        u_ref_n1_dfs = QueryBuilder().with_cals(ldb) \
            .with_duration(t_start=timestamp_qds, duration=duration) \
            .with_circuit_type(self.circuit_type) \
            .with_metadata(circuit_name=self.circuit_name, system='DIODE_RB', source=source_nqps,
                           wildcard={'QPS_CRATE': source_nqps}, signal='U_REF_N1') \
            .signal_query() \
            .synchronize_time(timestamp_qds / 1e9) \
            .dfs

        return u_diode_dfs, u_ref_n1_dfs

    def _query_nqps_pm(self, magnet_names, source_nqps, timestamp_nqps):

        u_diode_dfs = QueryBuilder().with_pm() \
            .with_timestamp(timestamp_nqps) \
            .with_circuit_type(self.circuit_type) \
            .with_metadata(circuit_name=self.circuit_name, system='DIODE_RB', source=source_nqps,
                           wildcard={'MAGNET': magnet_names}, signal='U_DIODE_RB') \
            .signal_query() \
            .synchronize_time(timestamp_nqps) \
            .convert_index_to_sec().dfs
        u_ref_n1_dfs = QueryBuilder().with_pm() \
            .with_timestamp(timestamp_nqps) \
            .with_circuit_type(self.circuit_type) \
            .with_metadata(circuit_name=self.circuit_name, system='DIODE_RB', source=source_nqps,
                           wildcard={'QPS_CRATE': source_nqps}, signal='U_REF_N1') \
            .signal_query() \
            .synchronize_time(timestamp_nqps) \
            .convert_index_to_sec().dfs

        return u_diode_dfs, u_ref_n1_dfs
