from typing import List

import pandas as pd

import lhcsmapi
from lhcsmapi.Time import Time
from lhcsmapi.analysis.CircuitQuery import CircuitQuery, get_quench_current, execution_count
from lhcsmapi.metadata.MappingMetadata import MappingMetadata
from lhcsmapi.pyedsl.QueryBuilder import QueryBuilder
from lhcsmapi.reference.Reference import Reference


class RqCircuitQuery(CircuitQuery):
    """Class  extending the CircuitQuery class with methods for creation of templates of analysis tables and query
    of voltage across magnets in main quadrupole circuits.

    """

    def create_report_analysis_template(self, source_timestamp_qds_df: pd.DataFrame,
                                        source_timestamp_nqps_df: pd.DataFrame, timestamp_fgc_rqd: int,
                                        timestamp_fgc_rqf: int, i_meas_rqd_df: pd.DataFrame, i_meas_rqf_df:pd.DataFrame,
                                        author: str) -> pd.DataFrame:
        """ Method creating a report analysis template for main quadrupole circuits

        :param source_timestamp_qds_df: table with QDS PM source and timestamp
        :type source_timestamp_qds_df: pd.DataFrame
        :param source_timestamp_nqps_df: table with QDS PM source and timestamp
        :type source_timestamp_nqps_df: pd.DataFrame
        :param timestamp_fgc_rqd: unix timestamp of an FGC event in RQD circuit in nanoseconds
        :type timestamp_fgc_rqd: int
        :param timestamp_fgc_rqf: unix timestamp of an FGC event in RQF circuit in nanoseconds
        :type timestamp_fgc_rqf: int
        :param i_meas_rqd_df: measured power converter current in RQD circuit
        :type i_meas_rqd_df: pd.DataFrame
        :param i_meas_rqf_df: measured power converter current in RQF circuit
        :type i_meas_rqf_df: pd.DataFrame
        :param author: NICE account name of the analysis author
        :type author: str
        :return: pre-filled pd.DataFrame with a template to be filled further with data from analysis
        of a powering event in main quadrupole (RQ) circuits
        """
        columns = Reference.read_quench_database_columns(circuit_type='RQ', table_type='LHCSM')
        report_table = pd.DataFrame(columns=columns)

        timestamp_fgc = min(timestamp_fgc_rqd, timestamp_fgc_rqf)

        report_table['magnet'] = source_timestamp_qds_df['source']
        report_table['circuit'] = '%s.%s' % (self.circuit_type, self.circuit_name.split('.')[1])
        report_table['nqps_crate']= report_table['magnet'].\
            apply(lambda col:MappingMetadata.get_crate_name_from_magnet_name(self.circuit_type, col))

        report_table['datetime_fgc'] = Time.to_string_short(timestamp_fgc)
        report_table['timestamp_iqps'] = source_timestamp_qds_df['timestamp']
        report_table['datetime_iqps'] = report_table['timestamp_iqps'].apply(lambda col: Time.to_string_short(col))
        report_table['timestamp_nqps'] = report_table['nqps_crate']\
            .apply(lambda col: source_timestamp_nqps_df.loc[source_timestamp_nqps_df['source'] == col,
                                                            'timestamp'].values[0])
        report_table['datetime_nqps'] = report_table['timestamp_nqps'].apply(lambda col: Time.to_string_short(col))
        timestamp_qds_0 = report_table.loc[0, 'timestamp_iqps']
        report_table['dt_quench'] = report_table['timestamp_iqps'].apply(lambda col: (col - timestamp_qds_0) * 1e-9)
        report_table['t_quench'] = report_table['timestamp_iqps'].apply(lambda col: (col - timestamp_fgc) * 1e-9)
        report_table['I_MEAS_RQD_quench'] = report_table['t_quench'].apply(lambda col:
                                                                           get_quench_current(i_meas_rqd_df, col))
        report_table['I_MEAS_RQF_quench'] = report_table['t_quench'].apply(lambda col:
                                                                           get_quench_current(i_meas_rqf_df, col))

        report_table['performed_by'] = author
        report_table['lhcsmapi_version'] = lhcsmapi.__version__
        report_table['nqps_db'] = report_table['t_quench'].apply(lambda col: 'PM' if col < 2 else 'NXCALS')
        report_table.drop(columns='t_quench', inplace=True)

        report_table['iqps_or_nqps_trigger'] = report_table. \
            apply(lambda col: 'nQPS' if col['timestamp_nqps'] < col['timestamp_iqps'] else 'iQPS', axis=1)

        return report_table

    @execution_count()
    def query_voltage_cals(self, system: str, signal: str, timestamp_fgc: int, *, ldb, duration=[(50, 's'), (150, 's')],
                           ) -> List[pd.DataFrame]:
        """ Method querying CALS for voltage signals (either voltage feelers or diodes).
        If voltage across diodes is queried, then in addition, there is a U_REF signal queried for those diodes
        without U_DIODE signal.

        :param system: name of a system to query according to metadata
        :type system: str
        :param signal: name of a signal to query
        :type signal: str
        :param timestamp_fgc: unix timestamp of an FGC event in a circuit in nanoseconds to which the query is synced
        :type timestamp_fgc: int
        :param ldb: CALS pytimber connector
        :type ldb: pytimber
        :param duration: search duration before and after the start time
        :type duration: List[Tuple(int, str)]
        :return: list of pd.DataFrames (if a signal was not queried, an empty pd.DataFrame is returned and warning
            containing the raw query is returned)
        """
        u_cals_dfs = super().query_voltage_cals(system, signal, timestamp_fgc, ldb=ldb, duration=duration)

        if 'DIODE' in system:
            rq_diode = MappingMetadata.read_magnet_to_cell_qps_crate(circuit_type='RQ')
            crates = list(rq_diode[rq_diode['Crate U_DIODE_RQx'].isna() & (rq_diode['Circuit'] == self.circuit_name)]
                          ['Crate U_REF_N1'].values)

            u_ref_dfs = QueryBuilder().with_cals(ldb) \
                .with_duration(t_start=int(timestamp_fgc), duration=duration) \
                .with_circuit_type(self.circuit_type) \
                .with_metadata(circuit_name=self.circuit_name, system=system, signal='U_REF_N1',
                               wildcard={'QPS_CRATE': crates}) \
                .signal_query() \
                .synchronize_time(timestamp_fgc / 1e9).dfs

            return u_cals_dfs + u_ref_dfs
        else:
            return u_cals_dfs
