import pandas as pd
from IPython.display import display_html
import matplotlib.pyplot as plt

from lhcsmapi.pyedsl.FeatureBuilder import FeatureBuilder
from lhcsmapi.dbsignal.ExpSignalAnalysis import ExpSignalAnalysis
from lhcsmapi.reference.Reference import Reference
from lhcsmapi.Time import Time


def get_df_col(dfs, col):
    for df in dfs:
        if col in df.columns[0]:
            return df, df.columns[0]
    return None, None


def calculate_resistance(u_i_hds_dfs):
    no_signal = int(len(u_i_hds_dfs) / 2)
    r_hds_dfs = []
    for index in range(1, no_signal + 1):
        u_hds_df, col_u_hds_df = get_df_col(u_i_hds_dfs, 'U_HDS_{}'.format(index))
        i_hds_df, col_i_hds_df = get_df_col(u_i_hds_dfs, 'I_HDS_{}'.format(index))

        r_value = u_hds_df.values / i_hds_df.values
        column = col_u_hds_df.replace('U', 'R')
        time = u_hds_df.index
        r_hds_dfs.append(pd.DataFrame(data=r_value, index=time, columns=[column]))
    return r_hds_dfs


def get_decay_start_index(hds_dfs):
    # find df with I_HDS_1
    i_hds_df, col_i_hds_df = get_df_col(hds_dfs, 'I_HDS_1')
    if i_hds_df is not None:
        decay_start_index = ExpSignalAnalysis.get_window_with_given_mean_and_std(i_hds_df[col_i_hds_df])
        if decay_start_index == 0:
            u_hds_df, col_u_hds_df = get_df_col(hds_dfs, 'U_HDS_1')
            if u_hds_df is not None:
                return ExpSignalAnalysis.get_window_with_deviation_from_initial_value(u_hds_df[col_u_hds_df])
            else:
                return 0
        else:
            return decay_start_index
    else:
        return 0


def get_decay_only(hds_dfs):
    decay_start_index = get_decay_start_index(hds_dfs)
    if decay_start_index == 0:
        return hds_dfs
    else:
        return [hds_df.iloc[(decay_start_index + 1):len(hds_df)] for hds_df in hds_dfs]


def analyze_single_qh(timestamp_qh, source_qh, u_hds_dfs, i_hds_dfs, u_hds_ref_dfs, i_hds_ref_dfs, circuit_type='RB'):

    # Get decay only
    u_i_hds_dfs = get_decay_only(u_hds_dfs + i_hds_dfs)
    u_i_hds_ref_dfs = get_decay_only(u_hds_ref_dfs + i_hds_ref_dfs)

    # Filter
    u_i_hds_dfs = [u_i_hds_df.rolling(window=3, min_periods=1).median() for u_i_hds_df in u_i_hds_dfs]
    u_i_hds_ref_dfs = [u_i_hds_ref_df.rolling(window=3, min_periods=1).median() for u_i_hds_ref_df in u_i_hds_ref_dfs]

    # Calculate resistance
    r_hds_dfs = calculate_resistance(u_i_hds_dfs)
    r_hds_ref_dfs = calculate_resistance(u_i_hds_ref_dfs)

    timestamp_ref = Reference.get_quench_heater_reference_discharge('RB', source_qh)

    # Calculate features
    # # compare to reference values
    first_last_u_df = FeatureBuilder().with_signal(u_i_hds_dfs) \
        .calculate_features(features=['first', 'last20mean'], index=timestamp_qh)

    first_last_u_comp_df = Reference.compare_features_to_reference(first_last_u_df, circuit_type, 'QH',
                                                                   wildcard={'CELL': source_qh})

    first_r_df = FeatureBuilder().with_signal(r_hds_dfs) \
        .calculate_features(features='first20mean', index=timestamp_qh)
    first_r_ref_df = FeatureBuilder().with_signal(r_hds_ref_dfs) \
        .calculate_features(features='first20mean', index=timestamp_qh)
    first_r_comp_df = Reference.compare_difference_of_features_to_reference(first_r_df, first_r_ref_df, circuit_type,
                                                                            'QH', wildcard={'CELL': source_qh})

    # # compare to the reference discharge
    tau_df = FeatureBuilder().with_signal(u_i_hds_dfs) \
        .calculate_features(features=['tau_charge'], index=timestamp_qh)
    tau_ref_df = FeatureBuilder().with_signal(u_i_hds_ref_dfs) \
        .calculate_features(features=['tau_charge'], index=timestamp_ref)

    tau_comp_df = Reference.compare_difference_of_features_to_reference(tau_df, tau_ref_df, circuit_type,
                                                                        'QH', wildcard={'CELL': source_qh})

    # # Plot signals
    plot_qh_discharge(timestamp_qh, source_qh, u_i_hds_dfs, r_hds_dfs, u_i_hds_ref_dfs, r_hds_ref_dfs)

    # Display comparison
    first_last_u_comp_styler = first_last_u_comp_df.style.set_table_attributes("style='display:inline'").set_caption(
        'Comparison of the inititial and final voltage')
    first_r_comp_styler = first_r_comp_df.style.set_table_attributes("style='display:inline'").set_caption(
        'Comparison of the initial resisitance to the reference')
    tau_comp_styler = tau_comp_df.style.set_table_attributes("style='display:inline'").set_caption(
        'Comparison of the discharge characteristic time to the reference')

    display_html(first_last_u_comp_styler._repr_html_() + first_r_comp_styler._repr_html_()
                 + tau_comp_styler._repr_html_(), raw=True)

    # Compare features
    is_first_last_u_correct = Reference.check_feature_comparison(first_last_u_comp_df)
    is_first_r_correct = Reference.check_feature_comparison(first_r_comp_df)
    is_tau_correct = Reference.check_feature_comparison(tau_comp_df)

    return is_first_last_u_correct and is_first_r_correct and is_tau_correct


def plot_qh_discharge(timestamp_qh, source_qh, u_i_hds_dfs, r_hds_dfs, u_i_hds_ref_dfs, r_hds_ref_dfs):
    no_signal = int(len(u_i_hds_dfs) / 2)

    fig, ax = plt.subplots(1, 3, figsize=(20, 7))
    for index in range(1, no_signal + 1):
        u_hds_df = get_df_col(u_i_hds_dfs, 'U_HDS_{}'.format(index))[0]
        u_hds_ref_df = get_df_col(u_i_hds_ref_dfs, 'U_HDS_{}'.format(index))[0]
        u_hds_df.plot(ax=ax[0])
        u_hds_ref_df.plot(ax=ax[0], style='--')

    ax[0].grid(True)
    ax[0].tick_params(labelsize=15)
    ax[0].set_xlabel("time, [s]", fontsize=15)
    ax[0].set_ylabel("U_HDS, [V]", fontsize=15)

    title = "Magnet: {}, Time Stamp: {}, U_HDS(t), I_HDS(t), and R_HDS(t)". \
        format(source_qh, Time.to_string_short(timestamp_qh))

    for index in range(1, no_signal + 1):
        i_hds_df = get_df_col(u_i_hds_dfs, 'I_HDS_{}'.format(index))[0]
        i_hds_ref_df = get_df_col(u_i_hds_ref_dfs, 'I_HDS_{}'.format(index))[0]
        i_hds_df.plot(ax=ax[1], title=title)
        i_hds_ref_df.plot(ax=ax[1], style='--')

    ax[1].title.set_size(20)
    ax[1].grid(True)
    ax[1].tick_params(labelsize=15)
    ax[1].set_xlabel("time, [s]", fontsize=15)
    ax[1].set_ylabel("I_HDS, [A]", fontsize=15)

    for r_hds_df, r_hds_ref_df in zip(r_hds_dfs, r_hds_ref_dfs):
        r_hds_df.plot(ax=ax[2])
        r_hds_ref_df.plot(ax=ax[2], style='--')
    ax[2].grid(True)
    ax[2].tick_params(labelsize=15)
    ax[2].set_xlabel("time, [s]", fontsize=15)
    ax[2].set_ylabel("R_HDS (Calculated), [Ohm]", fontsize=15)

    plt.show()