from typing import Tuple

import pandas as pd
from copy import deepcopy
import scipy.signal as spsig
import scipy as sp
import numpy as np

from lhcsmapi.dbsignal.SignalProcessing import SignalProcessing
from lhcsmapi.dbsignal.SignalAnalysis import SignalAnalysis
from lhcsmapi.dbsignal.SignalUtilities import SignalUtilities


def add_two_dataframes(df1, df2, col_name):
    idx = df1.index.union(df2.index)
    idx_df1 = df1.reindex(idx).interpolate('index')
    idx_df2 = df2.reindex(idx).interpolate('index')

    return pd.DataFrame(index=idx_df1.index, data=idx_df1.values + idx_df2.values, columns=[col_name])


def calculate_char_time_profile(i_meas_df, timestamp_first_quench=0):
    tau_i_meas_df = deepcopy(i_meas_df)
    signal = tau_i_meas_df.columns[0]
    der_name = 'd{}'.format(signal)
    SignalProcessing.calculate_time_derivative_of_signals(tau_i_meas_df, tau_i_meas_df.index, [signal], [der_name])

    # Calculate time constant as a ratio of a signal by its derivative
    char_time_name = '{0}/d{0}'.format(signal)
    tau_i_meas_df[char_time_name] = -tau_i_meas_df[signal] / sp.signal.medfilt(tau_i_meas_df[der_name], 3)
    tau_i_meas_df.loc[tau_i_meas_df.index < timestamp_first_quench, [char_time_name]] = np.nan
    tau_i_meas_df.drop(columns=[der_name], inplace=True)

    return tau_i_meas_df[char_time_name]


def calculate_time_derivative_of_signal(df: pd.DataFrame, kernel_size=51) -> pd.DataFrame:
    """ Function calculating time derivatives for signal contained in the passed DataFrame object.
    The returned signal name contains prefix d w.r.t. the input signal name, e.g. I_MEAS -> dI_MEAS.

    :param df: input DataFrame
    :param kernel_size: kernel size for gradient calculation
    :type kernel_size: int

    :return: pd.DataFrame with signal derivative
    """
    if isinstance(df, pd.DataFrame) and (df.shape[1] == 1) and not df.empty:
        signal_name = df.columns[0]
        der_name = 'd{}'.format(signal_name)
        return pd.DataFrame(spsig.medfilt(np.gradient(df[signal_name]) / np.gradient(df.index), kernel_size), df.index,
                            columns=[der_name])
    else:
        raise ValueError('Wrong input type, please check the input assumptions')


# ToDo: SignalTransformationBuilder
def flip_u_qs0_signal(u_qs0_ref: pd.DataFrame) -> pd.DataFrame:
    """ Function flipping the U_QS0 signal such that the first signal increase has positive slope.

    :param u_qs0_ref: U_QS0 QPS signal
    :type u_qs0_ref: pd.DataFrame

    :return: flipped U_QS0 signal
    """

    # # if there is only a negative peak, then flip the signal
    # # if there are two peaks and the first is negative, then flip the signal
    # # added the case when U_QS0 is saturated only to negative!
    # # otherwise, do nothing
    if all(u_qs0_ref.max() < 0.2) and all(u_qs0_ref.min() < -0.2):
        return -deepcopy(u_qs0_ref)
    elif (u_qs0_ref.max().values[0] > 0.2) & (u_qs0_ref.min().values[0] < -0.2) & (
            u_qs0_ref.idxmin().values[0] < u_qs0_ref.idxmax().values[0]):
        return -deepcopy(u_qs0_ref)
    else:
        return deepcopy(u_qs0_ref)


def calculate_last_const_di_dt(i_meas_df: pd.DataFrame, const_di_dt_duration=100) -> Tuple[float, float]:
    """ Function calculating the value and duration of the last di/dt prior to an event in the input signal.

    :param i_meas_df: pd.DataFrame with power converter current
    :type i_meas_df: pd.DataFrame
    :param const_di_dt_duration: minimum duration of constant di/dt in samples
    :type const_di_dt_duration: int

    :return: a tuple of last di/dt value and duration
    """
    # Calculate the first derivative
    d_i_meas_df = calculate_time_derivative_of_signal(i_meas_df[i_meas_df.index < 0], kernel_size=51)
    # To remove noise round the derivative to a single digit
    d_i_meas_df['round_dSTATUS.I_MEAS'] = d_i_meas_df.apply(lambda row: round(row['dSTATUS.I_MEAS'], 1), axis=1)
    # Calculate the 2nd derivative
    dd_i_meas_df = calculate_time_derivative_of_signal(pd.DataFrame(d_i_meas_df['round_dSTATUS.I_MEAS']), kernel_size=1)
    # Count duration of 2nd derivative equal to 0
    dd_i_meas_df['count'] = (dd_i_meas_df['dround_dSTATUS.I_MEAS'].diff(1) != 0).astype('int').cumsum()
    # Only take 2nd derivatives longer then a given number of points
    thresh = dd_i_meas_df['count'].value_counts() > const_di_dt_duration
    max_counts = thresh[thresh == True].index
    # Take indices of constant 2nd values
    index_max_counts = [dd_i_meas_df[dd_i_meas_df['count'] == max_count].index[0] for max_count in max_counts]

    # Sort according to the index_max_counts
    index_max_counts_sorted = sorted(index_max_counts, reverse=True)

    # Return the last constant didt
    last_di_dt = d_i_meas_df[dd_i_meas_df.index == max(index_max_counts_sorted)]['round_dSTATUS.I_MEAS'].values[0]
    index_max_count = index_max_counts_sorted.index(max(index_max_counts_sorted))
    if (index_max_count == 0) or (index_max_count == len(index_max_counts_sorted) - 1):
        last_di_dt_duration = -max(index_max_counts)
    else:
        last_di_dt_duration = index_max_counts_sorted[index_max_count] - index_max_counts_sorted[index_max_count + 1]

    return last_di_dt, last_di_dt_duration


def get_current_plateau_start_end(i_meas_df, i_meas_threshold=500):
    I_MEAS = i_meas_df.columns[0]
    i_meas_df['acqStamp'] = i_meas_df.index
    i_meas_df = SignalUtilities.synchronize_df(i_meas_df)
    i_meas_df = SignalUtilities.convert_indices_to_sec(i_meas_df)

    t0 = i_meas_df['acqStamp'].values[0]

    pd.options.mode.chained_assignment = None
    # take the current only above 500 A
    i_meas_df = i_meas_df[i_meas_df[I_MEAS] > i_meas_threshold]

    # calculate time derivative
    i_meas_df.loc[:, 'time'] = i_meas_df.index
    i_meas_df['dI_MEAS'] = i_meas_df[I_MEAS].diff()
    i_meas_df['dtime'] = i_meas_df['time'].diff()
    i_meas_df = i_meas_df.dropna()
    i_meas_df['dI_MEAS_dt'] = i_meas_df.apply(lambda col: col['dI_MEAS'] / col['dtime'], axis=1)
    # round to the nearest integer
    i_meas_df['round_dI_MEAS_dt'] = i_meas_df.apply(lambda row: np.around(row['dI_MEAS_dt']), axis=1)
    # get sign of the round value
    i_meas_df['sign_dI_MEAS_dt'] = i_meas_df.apply(lambda row: np.sign(row['round_dI_MEAS_dt']), axis=1)
    # find change of the sign of the first derivative
    i_meas_df['change_sign_dI_MEAS_dt'] = i_meas_df['sign_dI_MEAS_dt'].diff()
    di_meas_dt_change = i_meas_df[i_meas_df['change_sign_dI_MEAS_dt'] != 0]
    di_meas_dt_change.index = di_meas_dt_change['acqStamp']
    di_meas_dt_change.drop(columns=['acqStamp'], inplace=True)

    # Get time_start, time_end of plateaus
    pattern_ramp_up = [-1, 1]
    exact_ramp_up_df, _ = SignalAnalysis.find_rows_with_exact_pattern_in_series(
        di_meas_dt_change['change_sign_dI_MEAS_dt'], pattern_ramp_up)
    exact_ramp_up_df.rename(columns={-1: "plateau_start", 1: 'plateau_end'}, inplace=True)

    pattern_ramp_down = [-1, -1]
    exact_ramp_down_df, _ = SignalAnalysis.find_rows_with_exact_pattern_in_series(
        di_meas_dt_change['change_sign_dI_MEAS_dt'], pattern_ramp_down)
    exact_ramp_down_df.columns = ['plateau_start', 'plateau_end']

    plateau_timing = pd.concat([exact_ramp_up_df, exact_ramp_down_df])

    # ToDo: sort plateau_timing by plateau_start
    plateau_timing['plateau_start_sync'] = plateau_timing['plateau_start'].apply(lambda x: (x - t0) / 1e9)
    plateau_timing['plateau_end_sync'] = plateau_timing['plateau_end'].apply(lambda x: (x - t0) / 1e9)

    pd.options.mode.chained_assignment = 'warn'

    return plateau_timing
