import unittest
from unittest.mock import patch

import matplotlib.pyplot as plt

from lhcsmapi.analysis.busbar.BusbarResistanceAnalysis import find_plateau_start_and_end, \
    find_start_end_of_the_longest_ramp_up, extract_intersecting_plateaus
from test.resources.read_csv import read_csv


class TestBusbarResistanceAnalysis(unittest.TestCase):

    @patch("matplotlib.pyplot.show")
    def test_find_plateau_start_and_end(self, mock_show=None):
        # arrange
        i_meas_raw_df = read_csv('resources/busbar', 'I_MEAS')

        # act
        plateau_start_act, plateau_end_act = find_plateau_start_and_end(i_meas_raw_df, i_meas_threshold=50)

        ax = i_meas_raw_df.plot()
        for ps, pe in zip(plateau_start_act, plateau_end_act):
            ax.axvspan(xmin=ps, xmax=pe, facecolor='xkcd:green')
        plt.show()

        # assert
        plateau_start_ref = [1523190227160000000, 1523190778160000000, 1523193354660000000]
        plateau_end_ref = [1523190645160000000, 1523192221160000000, 1523195095160000000]

        self.assertListEqual(plateau_start_ref, plateau_start_act)
        self.assertListEqual(plateau_end_ref, plateau_end_act)

        if mock_show is not None:
            mock_show.assert_called()

    @patch("matplotlib.pyplot.show")
    def test_find_plateau_start_and_end_it_i_meas_rqx(self, mock_show=None):
        # arrange
        i_meas_raw_df = read_csv('resources/busbar', 'I_MEAS_RQX')

        # act
        plateau_start_act, plateau_end_act = find_plateau_start_and_end(i_meas_raw_df, i_meas_threshold=0,
                                                                        min_duration_in_sec=60)

        ax = i_meas_raw_df.plot()
        for ps, pe in zip(plateau_start_act, plateau_end_act):
            ax.axvspan(xmin=ps, xmax=pe, facecolor='xkcd:green')
        plt.show()

        # assert
        plateau_start_ref = [1491936032740000000, 1491937781240000000]
        plateau_end_ref = [1491937185240000000, 1491941327240000000]

        self.assertListEqual(plateau_start_ref, plateau_start_act)
        self.assertListEqual(plateau_end_ref, plateau_end_act)

        if mock_show is not None:
            mock_show.assert_called()

    @patch("matplotlib.pyplot.show")
    def test_find_plateau_start_and_end_it_i_meas_rtqx1(self, mock_show=None):
        # arrange
        i_meas_raw_df = read_csv('resources/busbar', 'I_MEAS_RTQX1')

        # act
        plateau_start_act, plateau_end_act = find_plateau_start_and_end(i_meas_raw_df, i_meas_threshold=0,
                                                                        min_duration_in_sec=60)

        ax = i_meas_raw_df.plot()
        for ps, pe in zip(plateau_start_act, plateau_end_act):
            ax.axvspan(xmin=ps, xmax=pe, facecolor='xkcd:green')
        plt.show()

        # assert
        plateau_start_ref = [1491935921240000000, 1491937377740000000]
        plateau_end_ref = [1491937189240000000, 1491941331240000000]

        self.assertListEqual(plateau_start_ref, plateau_start_act)
        self.assertListEqual(plateau_end_ref, plateau_end_act)

        if mock_show is not None:
            mock_show.assert_called()

    @patch("matplotlib.pyplot.show")
    def test_find_plateau_start_and_end_it_i_meas_rtqx2(self, mock_show=None):
        # arrange
        i_meas_raw_df = read_csv('resources/busbar', 'I_MEAS_RTQX2')

        # act
        plateau_start_act, plateau_end_act = find_plateau_start_and_end(i_meas_raw_df, i_meas_threshold=0,
                                                                        min_duration_in_sec=60)

        ax = i_meas_raw_df.plot()
        for ps, pe in zip(plateau_start_act, plateau_end_act):
            ax.axvspan(xmin=ps, xmax=pe, facecolor='xkcd:green')
        plt.show()

        # assert
        plateau_start_ref =[1491936040240000000, 1491937725240000000, 1491941715740000000]
        plateau_end_ref = [1491937185240000000, 1491941327240000000, 1491942256240000000]

        self.assertListEqual(plateau_start_ref, plateau_start_act)
        self.assertListEqual(plateau_end_ref, plateau_end_act)

        if mock_show is not None:
            mock_show.assert_called()

    @patch("matplotlib.pyplot.show")
    def test_find_plateau_start_and_end_it(self, mock_show=None):
        # arrange
        i_meas_raw_dfs = [read_csv('resources/busbar', 'I_MEAS_RQX'),
                          read_csv('resources/busbar', 'I_MEAS_RTQX1'),
                          read_csv('resources/busbar', 'I_MEAS_RTQX2')]

        # act
        plateau_starts = []
        plateau_ends = []
        for i_meas_raw_df in i_meas_raw_dfs:
            if i_meas_raw_df.max().values[0] > 100:
                plateau_start, plateau_end = find_plateau_start_and_end(i_meas_raw_df, i_meas_threshold=0,
                                                                        min_duration_in_sec=10)
                plateau_starts.append(plateau_start)
                plateau_ends.append(plateau_end)

        plateau_start_act, plateau_end_act = extract_intersecting_plateaus(plateau_ends, plateau_starts)
        for i_meas_raw_df in i_meas_raw_dfs:
            ax = i_meas_raw_df.plot(figsize=(15, 7))
            for ps, pe in zip(plateau_start_act, plateau_end_act):
                ax.axvspan(xmin=ps, xmax=pe, facecolor='xkcd:green')

        plt.show()

        # assert
        plateau_start_ref =[1491936040240000000, 1491937725240000000]
        plateau_end_ref = [1491937185240000000, 1491941327240000000]

        self.assertListEqual(plateau_start_ref, plateau_start_act)
        self.assertListEqual(plateau_end_ref, plateau_end_act)

        if mock_show is not None:
            mock_show.assert_called()

    @patch("matplotlib.pyplot.show")
    def test_find_start_end_of_the_longest_ramp_up(self, mock_show=None):
        # arrange
        i_meas_raw_df = read_csv('resources/busbar', 'I_MEAS')

        # act
        plateau_start, plateau_end = find_plateau_start_and_end(i_meas_raw_df, i_meas_threshold=50)

        ramp_up_start_act, ramp_up_end_act = find_start_end_of_the_longest_ramp_up(plateau_start, plateau_end)

        ax = i_meas_raw_df.plot()
        for ps, pe in zip(plateau_start, plateau_end):
            ax.axvspan(xmin=ps, xmax=pe, facecolor='xkcd:green')

        ax.axvspan(xmin=ramp_up_start_act, xmax=ramp_up_end_act, facecolor='xkcd:yellowgreen')
        plt.show()

        # assert
        ramp_up_start_ref = 1523192221160000000
        ramp_up_end_ref = 1523193354660000000

        self.assertEqual(ramp_up_start_ref, ramp_up_start_act)
        self.assertEqual(ramp_up_end_ref, ramp_up_end_act)

        if mock_show is not None:
            mock_show.assert_called()
