import unittest
import warnings
from unittest.mock import patch

import pandas as pd
import numpy as np

from lhcsmapi.analysis.R600ACircuitAnalysis import R600ACircuitAnalysis
from test.resources.read_csv import read_csv


class Test600ACircuitQuery(unittest.TestCase):

    # WITH/WITHOUT EEx
    def test_create_timestamp_table_missynchronized_qps_ee(self):
        # arrange
        timestamp_dct = {'FGC': 1493246176040000000,
                         'PIC': 1493246176022000000,
                         'EE': 1493246176049000000,
                         'QDS_A': 1493246176020000000,
                         'QDS_B': 1493246176021000000}
        analysis = R600ACircuitAnalysis('600A', pd.DataFrame(), is_automatic=True)

        # act
        with warnings.catch_warnings(record=True) as w:
            # Cause all warnings to always be triggered.
            warnings.simplefilter("always")
            # Trigger a warning.
            analysis.create_timestamp_table(timestamp_dct, 'ROD.A67B1')

            self.assertEquals('QDS_A and EE PM events are not synchronized to +/- 20 ms!', str(w[0].message))

    def test_create_timestamp_table_missynchronized_qps_fgc_ee(self):
        # arrange
        timestamp_dct = {'FGC': 1493246176040000000,
                         'PIC': 1493246176022000000,
                         'EE': 1493246176019000000,
                         'QDS_A': 1493246175924000000,
                         'QDS_B': 1493246175925000000}
        analysis = R600ACircuitAnalysis('600A', pd.DataFrame(), is_automatic=True)

        # act
        with warnings.catch_warnings(record=True) as w:
            # Cause all warnings to always be triggered.
            warnings.simplefilter("always")
            # Trigger a warning.
            analysis.create_timestamp_table(timestamp_dct, 'ROD.A67B1')

            self.assertEquals('QDS_A and FGC PM events are not synchronized to +/- 20 ms!', str(w[0].message))
            self.assertEquals('QDS_A and EE PM events are not synchronized to +/- 20 ms!', str(w[1].message))

    def test_create_timestamp_table_missing_qds_b_missynchronized_qps_fgc_ee(self):
        # arrange
        timestamp_dct = {'FGC': 1493246176040000000,
                         'PIC': 1493246176022000000,
                         'EE': 1493246176019000000,
                         'QDS_A': 1493246175924000000,
                         'QDS_B': np.nan}
        analysis = R600ACircuitAnalysis('600A', pd.DataFrame(), is_automatic=True)

        # act
        with warnings.catch_warnings(record=True) as w:
            # Cause all warnings to always be triggered.
            warnings.simplefilter("always")
            # Trigger a warning.
            analysis.create_timestamp_table(timestamp_dct, 'ROD.A67B1')

            self.assertEquals('QDS Board B not present in PM buffer!', str(w[0].message))
            self.assertEquals('QDS_A and FGC PM events are not synchronized to +/- 20 ms!', str(w[1].message))
            self.assertEquals('QDS_A and EE PM events are not synchronized to +/- 20 ms!', str(w[3].message))

    def test_find_time_of_quench(self):
        # arrange
        i_ref_df = read_csv('resources/hwc/600A', 'I_REF_ROD.A67B1')
        i_a_df = read_csv('resources/hwc/600A', 'I_A_ROD.A67B1')

        # act
        analysis = R600ACircuitAnalysis('600A', pd.DataFrame(), is_automatic=True)
        t_quench_act = analysis.find_time_of_quench(i_ref_df, i_a_df)

        # assert
        t_quench_ref = -0.079
        self.assertAlmostEqual(t_quench_ref, t_quench_act)

    @patch("lhcsmapi.pyedsl.PlotBuilder.plt.show")
    def test_plot_i_meas_pc(self, mock_show=None):
        # arrange
        i_meas_df = read_csv('resources/hwc/600A', 'I_MEAS_ROD.A67B1')
        i_ref_df = read_csv('resources/hwc/600A', 'I_REF_ROD.A67B1')
        i_a_df = read_csv('resources/hwc/600A', 'I_A_ROD.A67B1')
        circuit_name = 'ROD.A67B1'
        timestamp_fgc = 1493246176040000000

        # act
        analysis = R600ACircuitAnalysis('600A', pd.DataFrame(), is_automatic=True)
        t_quench = analysis.find_time_of_quench(i_ref_df, i_a_df)
        analysis.plot_i_meas_pc(circuit_name, timestamp_fgc, [i_meas_df, i_a_df], i_ref_df,
                                xlim=(t_quench-0.1, t_quench+0.1))

        # assert
        if mock_show is not None:
            mock_show.assert_called()

    def test_calculate_current_miits(self):
        # arrange
        i_meas_df = read_csv('resources/hwc/600A', 'I_MEAS_ROD.A67B1')
        i_ref_df = read_csv('resources/hwc/600A', 'I_REF_ROD.A67B1')
        i_a_df = read_csv('resources/hwc/600A', 'I_A_ROD.A67B1')

        # act
        analysis = R600ACircuitAnalysis('600A', pd.DataFrame(index=[0]), is_automatic=True)
        t_quench = analysis.find_time_of_quench(i_ref_df, i_a_df)
        analysis.calculate_current_miits(i_meas_df, t_quench)

        # assert
        self.assertAlmostEqual(0.002, analysis.results_table.loc[0, 'I_MEAS MIITs'], places=4)

    def test_calculate_quench_current(self):
        # arrange
        i_meas_df = read_csv('resources/hwc/600A', 'I_MEAS_ROD.A67B1')
        i_ref_df = read_csv('resources/hwc/600A', 'I_REF_ROD.A67B1')
        i_a_df = read_csv('resources/hwc/600A', 'I_A_ROD.A67B1')

        # act
        analysis = R600ACircuitAnalysis('600A', pd.DataFrame(index=[0]), is_automatic=True)
        t_quench = analysis.find_time_of_quench(i_ref_df, i_a_df)
        analysis.calculate_quench_current(i_meas_df, t_quench)

        # assert
        self.assertAlmostEqual(186.2, analysis.results_table.loc[0, 'Quench Current'], places=2)

    def test_calculate_current_slope(self):
        # arrange
        i_meas_df = read_csv('resources/hwc/600A', 'I_MEAS_ROD.A67B1')

        # act
        analysis = R600ACircuitAnalysis('600A', pd.DataFrame(index=[0]), is_automatic=True)
        analysis.calculate_current_slope(i_meas_df)

        # assert
        self.assertAlmostEqual(-0.7, analysis.results_table.loc[0, 'Ramp Rate'], places=2)

    @patch("lhcsmapi.pyedsl.PlotBuilder.plt.show")
    def test_plot_i_earth(self, mock_show=None):
        # arrange
        i_earth_df = read_csv('resources/hwc/600A', 'IEARTH.IEARTH_ROD.A67B1')
        circuit_name = 'ROD.A67B1'
        timestamp_fgc = 1493246176040000000

        # act
        analysis = R600ACircuitAnalysis('600A', pd.DataFrame(index=[0]), is_automatic=True)
        analysis.plot_i_earth_pc(circuit_name, timestamp_fgc, i_earth_df)

        # assert
        if mock_show is not None:
            mock_show.assert_called()

    def test_calculate_max_i_earth_pc(self):
        # arrange
        i_earth_df = read_csv('resources/hwc/600A', 'IEARTH.IEARTH_ROD.A67B1')

        # act
        analysis = R600ACircuitAnalysis('600A', pd.DataFrame(index=[0]), is_automatic=True)
        analysis.calculate_max_i_earth_pc(i_earth_df)

        # assert
        self.assertAlmostEqual(0.012, analysis.results_table.loc[0, 'Earth Current'], places=3)

    @patch("lhcsmapi.pyedsl.PlotBuilder.plt.show")
    def test_analyze_u_dump_res_ee(self, mock_show=None):
        # arrange
        i_meas_df = read_csv('resources/hwc/600A', 'I_MEAS_ROD.A67B1')
        u_dump_res_df = read_csv('resources/hwc/600A', 'U_DUMP_RES_ROD.A67B1')
        circuit_name = 'ROD.A67B1'
        timestamp_fgc = 1493246176040000000

        # act
        analysis = R600ACircuitAnalysis('600A', pd.DataFrame(index=[0]), is_automatic=True)
        analysis.analyze_u_dump_res_ee(circuit_name, timestamp_fgc, i_meas_df, u_dump_res_df)

        # assert
        self.assertAlmostEqual(113.57, analysis.results_table.loc[0, 'EE max voltage'], places=2)

        # assert
        if mock_show is not None:
            mock_show.assert_called()

    @patch("lhcsmapi.pyedsl.PlotBuilder.plt.show")
    def test_plot_u_res(self, mock_show=None):
        # arrange
        u_res_nxcals_df = read_csv('resources/hwc/600A', 'U_RES_NXCALS_ROD.A67B1')
        i_meas_df = read_csv('resources/hwc/600A', 'I_MEAS_ROD.A67B1')
        circuit_name = 'ROD.A67B1'
        timestamp_qds = 1493246175924000000

        # act
        analysis = R600ACircuitAnalysis('600A', pd.DataFrame(index=[0]), is_automatic=True)
        analysis.plot_u_res(circuit_name, timestamp_qds, u_res_nxcals_df, i_meas_df)

        # assert
        if mock_show is not None:
            mock_show.assert_called()

    def test_calculate_u_res_slope(self):
        # arrange
        u_res_df = read_csv('resources/hwc/600A', 'U_RES_ROD.A67B1')

        # act
        analysis = R600ACircuitAnalysis('600A', pd.DataFrame(index=[0]), is_automatic=True)
        u_res_slope_act_df = analysis.calculate_u_res_slope(u_res_df)

        # assert
        u_res_slope_ref_df = pd.DataFrame({'U_RES_slope': {0.062: 0.023349186, 0.104: 0.20909318}})

        pd.testing.assert_frame_equal(u_res_slope_ref_df, u_res_slope_act_df)
        self.assertAlmostEqual(4.422, analysis.results_table.loc[0, 'dU_RES_dt'], places=2)

    @patch("lhcsmapi.pyedsl.PlotBuilder.plt.show")
    def test_plot_u_res_slope(self, mock_show=None):
        # arrange
        u_res_df = read_csv('resources/hwc/600A', 'U_RES_ROD.A67B1')
        circuit_name = 'ROD.A67B1'
        timestamp_qds = 1493246175924000000

        # act
        analysis = R600ACircuitAnalysis('600A', pd.DataFrame(index=[0]), is_automatic=True)
        u_res_slope_df = analysis.calculate_u_res_slope(u_res_df)
        analysis.plot_u_res_slope(circuit_name, timestamp_qds, u_res_df, u_res_slope_df)

        # assert
        if mock_show is not None:
            mock_show.assert_called()

    @patch("lhcsmapi.pyedsl.PlotBuilder.plt.show")
    def test_plot_qds(self, mock_show=None):
        # arrange
        u_res_df = read_csv('resources/hwc/600A', 'U_RES_ROD.A67B1')
        u_diff_df = read_csv('resources/hwc/600A', 'U_DIFF_ROD.A67B1')
        i_didt_df = read_csv('resources/hwc/600A', 'I_DIDT_ROD.A67B1')
        i_dcct_df = read_csv('resources/hwc/600A', 'I_DCCT_ROD.A67B1')
        circuit_name = 'ROD.A67B1'
        timestamp_qds = 1493246175924000000

        # act
        analysis = R600ACircuitAnalysis('600A', pd.DataFrame(index=[0]), is_automatic=True)
        analysis.plot_qds(circuit_name, timestamp_qds, i_dcct_df, i_didt_df, u_diff_df, u_res_df)

        # assert
        if mock_show is not None:
            mock_show.assert_called()

    def test_create_mp3_results_table(self):
        # arrange
        circuit_type = '600A'
        is_automatic = True
        results_table = pd.DataFrame({'Circuit Name': {0: 'ROD.A67B1'}, 'Family': {0: 'ROD'}, 'Period': {0: np.nan},
                                      'Date': {0: '2017-04-27'}, 'Time': {0: '00:36:16.040'},
                                      'flagged in APEX': {0: np.nan}, 'Quench Current': {0: 185.6},
                                      'I_MEAS MIITs': {0: 0.018}, 'Reason': {0: np.nan}, 'Pivot': {0: np.nan},
                                      'Quench Origin': {0: np.nan}, 'Ramp Rate': {0: -0.7},
                                      'Plateau Duration': {0: np.nan}, 'dUres/dt': {0: 4.422476047619048},
                                      'EE max voltage': {0: 113.57}, 'Earth Current': {0: 0.012},
                                      'Coupling': {0: np.nan}, 'Comment': {0: np.nan}, 'performed_by': {0: 'mmacieje'},
                                      'Reference Doc': {0: np.nan}, 'lhcsmapi_version': {0: '1.3.226'},
                                      'timestamp_sync': {0: 1493246176040000000},
                                      'timestamp_pic': {0: 1493246176022000000},
                                      'timestamp_ee': {0: 1493246176019000000},
                                      'timestamp_qds_a': {0: 1493246175924000000},
                                      'ROD.A67B1_LEADS_U_HTS': {0: 'Within reference'},
                                      'ROD.A67B1_LEADS_U_RES': {0: 'Within reference'}})
        analysis_rcd = R600ACircuitAnalysis(circuit_type, results_table, is_automatic=is_automatic)

        # act
        mp3_act = analysis_rcd.create_mp3_results_table()

        # assert
        mp3_ref = pd.DataFrame({'Circuit Name': {0: 'ROD.A67B1'}, 'Family': {0: 'ROD'}, 'Period': {0: np.nan},
                                'Date': {0: '2017-04-27'}, 'Time': {0: '00:36:16.040'}, 'flagged in APEX': {0: np.nan},
                                'Quench Current': {0: 185.6}, 'DCCT MIITs': {0: 0.018}, 'Reason': {0: np.nan},
                                'Pivot': {0: np.nan}, 'Quench Origin': {0: np.nan}, 'Ramp Rate': {0: -0.7},
                                'Plateau Duration': {0: np.nan}, 'dUres/dt': {0: 4.422476047619048},
                                'EE max voltage': {0: 113.57}, 'Earth Current': {0: 0.012}, 'Coupling': {0: np.nan},
                                'Comment': {0: np.nan}, 'Person who analyzed': {0: 'mmacieje'},
                                'Reference Doc': {0: np.nan}})

        mp3_act.fillna(np.nan, inplace=True)
        pd.testing.assert_frame_equal(mp3_ref, mp3_act)

    def test_create_mp3_results_table_rcdo(self):
        # arrange
        circuit_type = '600A'
        is_automatic = True
        results_table = pd.DataFrame({'Circuit Name': {0: 'RCD/O.A12B2'}, 'Family': {0: 'RCD'}, 'Period': {0: np.nan},
                                      'Date': {0: '2017-04-25'}, 'Time': {0: '18:55:03.160'},
                                      'flagged in APEX': {0: np.nan}, 'I_Q_RCO': {0: -105.0}, 'I_Q_RCD': {0: 349.7},
                                      'I_MEAS MIITs RCO': {0: 0.0}, 'I_MEAS MIITs RCD': {0: 0.004},
                                      'Reason': {0: np.nan}, 'Quench Origin': {0: np.nan},
                                      'Ramp Rate RCO': {0: 0.0}, 'Ramp Rate RCD': {0: 1.5},
                                      'Plateau Duration RCO': {0: 148.72}, 'Plateau Duration RCD': {0: 148.72},
                                      'dUres/dt RCO': {0: -97.546145}, 'dUres/dt RCD': {0: 6.9692185625},
                                      'Comment': {0: np.nan}, 'performed by': {0: np.nan}, 'Reference Doc': {0: np.nan},
                                      'performed_by': {0: ''}, 'lhcsmapi_version': {0: '1.3.226'},
                                      'timestamp_fgc_rcd': {0: 1493139303220000000},
                                      'timestamp_fgc_rco': {0: 1493139303160000000},
                                      'timestamp_pic_rcd': {0: 1493139303213000000},
                                      'timestamp_pic_rco': {0: 1493139303142000000},
                                      'timestamp_qds_a_rcd': {0: 1493139303215000000},
                                      'timestamp_qds_a_rco': {0: 1493139303146000000},
                                      'Earth Current RCD': {0: -0.062}, 'Earth Current RCO': {0: 0.008},
                                      'RCD.A12B2_LEADS_U_HTS': {0: 'Within reference'},
                                      'RCD.A12B2_LEADS_U_RES': {0: 'Within reference'},
                                      'RCO.A12B2_LEADS_U_HTS': {0: 'Within reference'},
                                      'RCO.A12B2_LEADS_U_RES': {0: 'Within reference'}})
        analysis_rcd = R600ACircuitAnalysis(circuit_type, results_table, is_automatic=is_automatic)

        # act
        mp3_act = analysis_rcd.create_mp3_results_table_rcdo()

        # assert
        mp3_ref = pd.DataFrame({'Circuit Name': {0: 'RCD/O.A12B2'}, 'Family': {0: 'RCD'}, 'Period': {0: np.nan},
                                'Date': {0: '2017-04-25'}, 'Time': {0: '18:55:03.160'}, 'flagged in APEX': {0: np.nan},
                                'I_Q_RCO': {0: -105.0}, 'I_Q_RCD': {0: 349.7}, 'I_MEAS MIITs RCO': {0: 0.0},
                                'I_MEAS MIITs RCD': {0: 0.004}, 'Reason': {0: np.nan}, 'Quench Origin': {0: np.nan},
                                'Ramp Rate RCO': {0: 0.0}, 'Ramp Rate RCD': {0: 1.5},
                                'Plateau Duration RCO': {0: 148.72}, 'Plateau Duration RCD': {0: 148.72},
                                'dUres/dt RCO': {0: -97.546145}, 'dUres/dt RCD': {0: 6.9692185625},
                                'Comment': {0: np.nan}, 'Person who analyzed': {0: ''}, 'Reference Doc': {0: np.nan}})

        mp3_act.fillna(np.nan, inplace=True)
        pd.testing.assert_frame_equal(mp3_ref, mp3_act)

    def test_create_mp3_results_table_rcbxhv(self):
        # arrange
        circuit_type = '600A'
        is_automatic = True
        results_table = pd.DataFrame({'Circuit Name': {0: 'RCBXV1.R1'}, 'Family': {0: 'MCBX'},
                                      'Period': {0: np.nan}, 'Date': {0: '2017-04-26'},
                                      'Time': {0: '17:07:49.220'}, 'flagged in APEX': {0: np.nan},
                                      'I_Q_H': {0: 134.0}, 'I_Q_V': {0: 387.6}, 'radius': {0: np.nan},
                                      'I_MEAS MIITs H': {0: 0.018}, 'I_MEAS MIITs V': {0: 0.084},
                                      'Reason': {0: np.nan}, 'Quench Origin': {0: np.nan},
                                      'Ramp Rate H': {0: 2.2}, 'Ramp Rate V': {0: -0.7},
                                      'Plateau Duration H': {0: np.nan}, 'Plateau Duration V': {0: np.nan},
                                      'dUres/dt H': {0: 87.70904875000001}, 'dUres/dt V': {0: np.nan},
                                      'Comment': {0: np.nan}, 'performed by': {0: np.nan},
                                      'Reference Doc': {0: np.nan}, 'performed_by': {0: ''},
                                      'lhcsmapi_version': {0: '1.3.226'},
                                      'timestamp_fgc_rcbxh': {0: 1493219269240000000},
                                      'timestamp_fgc_rcbxv': {0: 1493219269220000000},
                                      'timestamp_pic_rcbxh': {0: 1493219269239000000},
                                      'timestamp_pic_rcbxv': {0: 1493219269212000000},
                                      'timestamp_qds_a_rcbxh': {0: 1493219269213000000},
                                      'Earth Current H': {0: 0.004}, 'Earth Current V': {0: -0.006},
                                      'RCBXH1.R1_LEADS_U_HTS': {0: 'Within reference'},
                                      'RCBXH1.R1_LEADS_U_RES': {0: 'Within reference'},
                                      'RCBXV1.R1_LEADS_U_HTS': {0: 'Within reference'},
                                      'RCBXV1.R1_LEADS_U_RES': {0: 'Outside of reference'}})
        analysis_rcbxh = R600ACircuitAnalysis(circuit_type, results_table, is_automatic=is_automatic)

        # act
        mp3_act = analysis_rcbxh.create_mp3_results_table_rcbxhv()

        # assert
        mp3_ref = pd.DataFrame({'Circuit Name': {0: 'RCBXV1.R1'}, 'Family': {0: 'MCBX'}, 'Period': {0: np.nan},
                                'Date': {0: '2017-04-26'}, 'Time': {0: '17:07:49.220'}, 'flagged in APEX': {0: np.nan},
                                'I_Q_H': {0: 134.0}, 'I_Q_V': {0: 387.6}, 'radius': {0: 410.1094488060474},
                                'I_MEAS MIITs H': {0: 0.018}, 'I_MEAS MIITs V': {0: 0.084}, 'Reason': {0: np.nan},
                                'Quench Origin': {0: np.nan}, 'Ramp Rate H': {0: 2.2}, 'Ramp Rate V': {0: -0.7},
                                'Plateau Duration H': {0: np.nan}, 'Plateau Duration V': {0: np.nan},
                                'dUres/dt H': {0: 87.70904875000001}, 'dUres/dt V': {0: np.nan}, 'Comment': {0: np.nan},
                                'Person who analyzed': {0: ''},
                                'Reference Doc': {0: np.nan}})

        mp3_act.fillna(np.nan, inplace=True)
        pd.testing.assert_frame_equal(mp3_ref, mp3_act)
