import unittest
import warnings
from io import StringIO
from unittest.mock import patch

import pandas as pd
import numpy as np

from lhcsmapi.analysis.RbCircuitAnalysis import RbCircuitAnalysis
from lhcsmapi.analysis.diode.DiodeLeadResistanceAnalysis import DiodeLeadResistanceAnalysis
from lhcsmapi.analysis.qds.QdsRbAnalysis import analyze_qds_trigger
from lhcsmapi.pyedsl.AssertionBuilder import AssertionBuilder
from test.test_analysis.hwc.read_csv import read_csv

import matplotlib as mpl
mpl.rc('figure', max_open_warning = 0)


class RbCircuitAnalysisTest(unittest.TestCase):

    def setUp(self):
        self.circuit_type = 'RB'
        self.circuit_name = 'RB.A12'
        self.timestamp_fgc = 1544631694840000000
        self.timestamp_fgc_ref = 1521298828680000000
        self.timestamp_pic = 1544631694800000000
        self.timestamp_ee_odd = 1544631694896000000
        self.timestamp_ee_even = 1544631695395000000
        results_table = read_csv('rb', 'fpa', 'results_table')
        results_table = results_table[results_table.index == 0]
        self.rb_analysis = RbCircuitAnalysis(self.circuit_type, results_table, True)

    def test_analyze_pic_equal(self):
        # arrange
        timestamp_pic = [1544631694800000000, 1544631694800000000]

        # act
        # assert
        output_ref = 'PIC timestamps (2018-12-12 17:21:34.800) and (2018-12-12 17:21:34.800) are less then 1 ms away.' \
                     '\nOUTPUT' \
                     '\n------' \
                     '\nPIC Analysis comment is: Automatic analysis, no expert feedback.' \
                     '\n------'
        with patch('sys.stdout', new=StringIO()) as output_act:
            self.rb_analysis.analyze_pic(timestamp_pic)
            self.assertEqual(output_ref, output_act.getvalue().strip())

    def test_analyze_pic_non_equal(self):
        # arrange
        timestamp_pic = [1544631694800000000, 1544631694802000000]

        # act
        # assert
        warning_ref = 'PIC timestamps (2018-12-12 17:21:34.800) and (2018-12-12 17:21:34.802) differ by more than ' \
                      '1 ms (2000000 ns)'
        output_ref = 'OUTPUT' \
                     '\n------' \
                     '\nPIC Analysis comment is: Automatic analysis, no expert feedback.' \
                     '\n------'
        with warnings.catch_warnings(record=True) as w,\
                patch('sys.stdout', new=StringIO()) as output_act:
            self.rb_analysis.analyze_pic(timestamp_pic)
            self.assertEqual(output_ref, output_act.getvalue().strip())
            self.assertEqual(warning_ref, str(w[0].message))

    def test_analyze_pic_nan(self):
        # arrange
        timestamp_pic = [1544631694800000000, np.nan]

        # act
        # assert
        warning_ref = 'At least one timestamp is NaN, check of timestamp PIC difference is skipped.'
        output_ref = 'OUTPUT' \
                     '\n------' \
                     '\nPIC Analysis comment is: Automatic analysis, no expert feedback.' \
                     '\n------'
        with warnings.catch_warnings(record=True) as w,\
                patch('sys.stdout', new=StringIO()) as output_act:
            self.rb_analysis.analyze_pic(timestamp_pic)
            self.assertEqual(output_ref, output_act.getvalue().strip())
            self.assertEqual(warning_ref, str(w[0].message))

    @patch("lhcsmapi.pyedsl.PlotBuilder.plt.show")
    def test_analyze_i_meas_pc(self, mock_show=None):
        # arrange
        i_meas_df = read_csv('rb', 'fpa', 'i_meas_df')
        i_meas_ref_df = read_csv('rb', 'fpa', 'i_meas_ref_df')

        # act
        self.rb_analysis.analyze_i_meas_pc(self.circuit_name, self.timestamp_fgc, self.timestamp_fgc_ref,
                                           self.timestamp_pic, i_meas_df, i_meas_ref_df)

        # assert
        if mock_show is not None:
            mock_show.assert_called()

    @patch("lhcsmapi.pyedsl.PlotBuilder.plt.show")
    def test_analyze_i_earth_pc(self, mock_show=None):
        # arrange
        i_a_df = read_csv('rb', 'fpa', 'i_a_df')
        i_earth_df = read_csv('rb', 'fpa', 'i_earth_df')
        i_earth_ref_df = read_csv('rb', 'fpa', 'i_earth_ref_df')

        # act
        self.rb_analysis.analyze_i_earth_pc(self.circuit_type, self.circuit_name, self.timestamp_fgc, i_a_df,
                                            i_earth_df, i_earth_ref_df)

        # assert
        if mock_show is not None:
            mock_show.assert_called()

    @patch("lhcsmapi.pyedsl.PlotBuilder.plt.show")
    def test_analyze_i_earth_pcnt_pc(self, mock_show=None):
        # arrange
        i_meas_df = read_csv('rb', 'fpa', 'i_meas_df')
        i_meas_ref_df = read_csv('rb', 'fpa', 'i_meas_ref_df')
        i_earth_pcnt_df = read_csv('rb', 'fpa', 'i_earth_pcnt_df')
        i_earth_pcnt_ref_df = read_csv('rb', 'fpa', 'i_earth_pcnt_ref_df')

        # act
        self.rb_analysis.analyze_i_earth_pcnt_pc(self.circuit_type, self.circuit_name, self.timestamp_fgc, i_meas_df,
                                                 i_meas_ref_df, i_earth_pcnt_df, i_earth_pcnt_ref_df)

        # assert
        if mock_show is not None:
            mock_show.assert_called()

    @patch("lhcsmapi.pyedsl.PlotBuilder.plt.show")
    def test_analyze_char_time_u_dump_res_ee_odd(self, mock_show=None):
        # arrange
        u_dump_res_odd_df = read_csv('rb', 'fpa', 'u_dump_res_odd_df')
        i_meas_df = read_csv('rb', 'fpa', 'i_meas_df')

        # act
        self.rb_analysis.analyze_char_time_u_dump_res_ee(self.circuit_name, self.timestamp_fgc, u_dump_res_odd_df,
                                                         i_meas_df)

        # assert
        if mock_show is not None:
            mock_show.assert_called()

    @patch("lhcsmapi.pyedsl.PlotBuilder.plt.show")
    def test_analyze_char_time_u_dump_res_ee_odd_even(self, mock_show=None):
        # arrange
        u_dump_res_odd_df = read_csv('rb', 'fpa', 'u_dump_res_odd_df')
        u_dump_res_even_df = read_csv('rb', 'fpa', 'u_dump_res_even_df')
        i_meas_df = read_csv('rb', 'fpa', 'i_meas_df')

        # act
        self.rb_analysis.analyze_char_time_u_dump_res_ee(self.circuit_name, self.timestamp_fgc,
                                                         [u_dump_res_odd_df, u_dump_res_even_df], i_meas_df)

        # assert
        if mock_show is not None:
            mock_show.assert_called()

    @patch("lhcsmapi.pyedsl.PlotBuilder.plt.show")
    def test_analyze_delay_time_u_dump_res_ee(self, mock_show=None):
        # arrange
        i_a_df = read_csv('rb', 'fpa', 'i_a_df')
        i_ref_df = read_csv('rb', 'fpa', 'i_ref_df')
        u_dump_res_odd_df = read_csv('rb', 'fpa', 'u_dump_res_odd_df')
        u_dump_res_even_df = read_csv('rb', 'fpa', 'u_dump_res_even_df')

        # act
        self.rb_analysis.analyze_delay_time_u_dump_res_ee(self.circuit_name, self.timestamp_fgc, self.timestamp_pic,
                                                          [self.timestamp_ee_odd, self.timestamp_ee_even], i_a_df,
                                                          i_ref_df, [u_dump_res_odd_df, u_dump_res_even_df])

        # assert
        if mock_show is not None:
            mock_show.assert_called()

    def test_temperature_ee_odd(self):
        # arrange
        t_res_body_odd_dfs = [read_csv('rb', 'fpa', 't_res_body_1_odd_0_df'),
                              read_csv('rb', 'fpa', 't_res_body_2_odd_0_df'),
                              read_csv('rb', 'fpa', 't_res_body_3_odd_0_df'),
                              read_csv('rb', 'fpa', 't_res_body_1_odd_1_df'),
                              read_csv('rb', 'fpa', 't_res_body_2_odd_1_df'),
                              read_csv('rb', 'fpa', 't_res_body_3_odd_1_df')]


        t_res_body_odd_ref_dfs = [read_csv('rb', 'fpa', 't_res_body_1_odd_0_ref_df'),
                                  read_csv('rb', 'fpa', 't_res_body_2_odd_0_ref_df'),
                                  read_csv('rb', 'fpa', 't_res_body_3_odd_0_ref_df'),
                                  read_csv('rb', 'fpa', 't_res_body_1_odd_1_ref_df'),
                                  read_csv('rb', 'fpa', 't_res_body_2_odd_1_ref_df'),
                                  read_csv('rb', 'fpa', 't_res_body_3_odd_1_ref_df')]
        # act
        AssertionBuilder().with_signal(t_res_body_odd_dfs) \
            .compare_to_reference(signal_ref_dfs=t_res_body_odd_ref_dfs, abs_margin=25, scaling=1)
        
    def test_temperature_ee_even(self):
        # arrange
        t_res_body_even_dfs = [read_csv('rb', 'fpa', 't_res_body_1_even_0_df'),
                               read_csv('rb', 'fpa', 't_res_body_2_even_0_df'),
                               read_csv('rb', 'fpa', 't_res_body_3_even_0_df'),
                               read_csv('rb', 'fpa', 't_res_body_1_even_1_df'),
                               read_csv('rb', 'fpa', 't_res_body_2_even_1_df'),
                               read_csv('rb', 'fpa', 't_res_body_3_even_1_df')]

        t_res_body_even_ref_dfs = [read_csv('rb', 'fpa', 't_res_body_1_even_0_ref_df'),
                                   read_csv('rb', 'fpa', 't_res_body_2_even_0_ref_df'),
                                   read_csv('rb', 'fpa', 't_res_body_3_even_0_ref_df'),
                                   read_csv('rb', 'fpa', 't_res_body_1_even_1_ref_df'),
                                   read_csv('rb', 'fpa', 't_res_body_2_even_1_ref_df'),
                                   read_csv('rb', 'fpa', 't_res_body_3_even_1_ref_df')]
        # act
        AssertionBuilder().with_signal(t_res_body_even_dfs) \
            .compare_to_reference(signal_ref_dfs=t_res_body_even_ref_dfs, abs_margin=25, scaling=1)

    @patch("lhcsmapi.pyedsl.PlotBuilder.plt.show")
    def test_analyze_qh(self, mock_show=None):
        # arrange
        u_hds_dfs = [[read_csv('rb', 'fpa', 'u_hds_1_df'), read_csv('rb', 'fpa', 'u_hds_2_df'),
                      read_csv('rb', 'fpa', 'u_hds_3_df'), read_csv('rb', 'fpa', 'u_hds_4_df')]]
        i_hds_dfs = [[read_csv('rb', 'fpa', 'i_hds_1_df'), read_csv('rb', 'fpa', 'i_hds_2_df'),
                      read_csv('rb', 'fpa', 'i_hds_3_df'), read_csv('rb', 'fpa', 'i_hds_4_df')]]

        u_hds_ref_dfs = [[read_csv('rb', 'fpa', 'u_hds_1_ref_df'), read_csv('rb', 'fpa', 'u_hds_2_ref_df'),
                          read_csv('rb', 'fpa', 'u_hds_3_ref_df'), read_csv('rb', 'fpa', 'u_hds_4_ref_df')]]
        i_hds_ref_dfs = [[read_csv('rb', 'fpa', 'i_hds_1_ref_df'), read_csv('rb', 'fpa', 'i_hds_2_ref_df'),
                          read_csv('rb', 'fpa', 'i_hds_3_ref_df'), read_csv('rb', 'fpa', 'i_hds_4_ref_df')]]
        # act
        self.rb_analysis.analyze_qh(u_hds_dfs, i_hds_dfs, u_hds_ref_dfs, i_hds_ref_dfs)

        # assert
        if mock_show is not None:
            mock_show.assert_called()

    def test_analyze_qds_trigger(self):
        # arrange
        u_qs0_df = read_csv('rb', 'fpa', 'u_qs0_0_df')
        st_nqd0_df = read_csv('rb', 'fpa', 'st_nqd0_0_df')
        st_magnet_ok_df = read_csv('rb', 'fpa', 'st_magnet_ok_0_df')

        # act
        report_qds_trigger_act = analyze_qds_trigger(u_qs0_df, st_magnet_ok_df, st_nqd0_df)

        # assert
        report_qds_trigger_ref = pd.DataFrame({'t_st_nqd0 [s]': {0: -0.026},
                                               't_st_magnet_ok [s]': {0: -0.016},
                                               'u_st_nqd0 [V]': {0: 0.0019524792},
                                               't_start_quench [s]': {0: -0.038},
                                               'du_dt [V/s]': {0: -7.914204933333333},
                                               't_delay_qh_trigger [s]': {0: 0.016},
                                               'analysis_comment': {0: ''}})

        pd.testing.assert_frame_equal(report_qds_trigger_ref, report_qds_trigger_act)

    def test_create_mp3_results_table(self):
        # arrange
        is_automatic = True
        results_table_final = read_csv('rb', 'fpa', 'results_table_final')

        # act
        rb_analysis = RbCircuitAnalysis(self.circuit_type, results_table_final, is_automatic=is_automatic)
        mp3_results_table_act = rb_analysis.create_mp3_results_table(self.circuit_name, self.timestamp_fgc)
        mp3_results_table_act.fillna(np.nan, inplace=True)

        # assert
        mp3_results_table_ref = read_csv('rb', 'fpa', 'mp3_results_table')
        mp3_results_table_ref.fillna(np.nan, inplace=True)

        pd.testing.assert_frame_equal(mp3_results_table_ref, mp3_results_table_act)

    # PLI3.a2
    @patch("lhcsmapi.pyedsl.PlotBuilder.plt.show")
    def test_analyze_delay_time_u_dump_res_ee_hwc(self, mock_show=None):
        # arrange
        circuit_name = 'RB.A12'
        timestamp_fgc = 1418473597620000000
        timestamp_ee_odd = 1418473749786000000
        timestamp_ee_even = 1418473750032000000

        i_a_df = read_csv('rb', 'pli3_d2', 'I_A')
        i_ref_df = read_csv('rb', 'pli3_d2', 'I_REF')
        u_dump_res_odd_df = read_csv('rb', 'pli3_d2', 'U_DUMP_RES_ODD')
        u_dump_res_even_df = read_csv('rb', 'pli3_d2', 'U_DUMP_RES_EVEN')

        is_automatic = True
        results_table = None

        # act
        rb_analysis = RbCircuitAnalysis(self.circuit_type, results_table, is_automatic=is_automatic)
        rb_analysis.analyze_delay_time_u_dump_res_ee(circuit_name, timestamp_fgc, timestamp_fgc,
                                                     [timestamp_ee_odd, timestamp_ee_even],
                                                     i_a_df, i_ref_df, [u_dump_res_odd_df, u_dump_res_even_df])

        # assert
        if mock_show is not None:
            mock_show.assert_called()
